import os
import gzip
import shutil
import logging
import requests
from urllib.parse import urlparse
from . import acquisition

l = logging.getLogger("pandem.zenodo")

class AcquisitionZenodo(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings):
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "zenodo")
    
    def new_files(self, dls, last_hash):
        source = dls['acquisition']['channel']['search']
        match = dls['acquisition']['channel']['match']
        url = f'https://zenodo.org/api/records/?q=conceptrecid:{source}&access_token='
        r = requests.get(url)
        response_status(r.status_code, url=url)

        j = r.json()
        if len(j['hits']['hits']) != 1:
            raise ValueError(f'Cannot identify a unique source with the provided string')
        hashes = last_hash.split(';')
        pairs = [(f['links']['self'], f['checksum']) for f in j['hits']['hits'][0]['files'] if f['checksum'] not in hashes]
        current_hash = []
        files_to_pipeline = []
        for url, file_hash in pairs:
            parts = urlparse(url)
            dest_name = parts.path.replace('/', '_')
            file_path = self.source_path(dls, dest_name)
            
            r = requests.get(url)
            response_status(r.status_code, url=url)
            
            current_hash.append(file_hash)
            if not os.path.exists(file_path[:-3]) or last_hash == "":
                with open(file_path, 'wb') as cont:
                    cont.write(r.content)
                try:
                    with gzip.open(file_path, 'rb') as f_in:
                        with open(file_path[:-3], 'wb') as f_out:
                            shutil.copyfileobj(f_in, f_out)
                    os.remove(file_path)
                    fpath_no_ext = os.path.splitext(file_path)[0]
                    files_to_pipeline.extend([fpath_no_ext])
                except gzip.BadGzipFile:
                    l.debug(f'{file_path} is corrupted. Skipping it')
        return {"hash":";".join(current_hash), "files":files_to_pipeline}


def response_status(rstatus: int, url: str = 'URL') -> None:
    if 200 <= rstatus and rstatus < 300:
        l.debug(f'{url} was accessed successfully')
    elif 500 <= rstatus and rstatus < 600:
        raise ValueError(f'Zenodo API is currently unavailable: {rstatus}')
    else:
        raise ValueError(f'{url} returns ERROR CODE {rstatus}')


import pandas as pd
import pkg_resources
import json
import codecs
import threading
import os
import shutil
import subprocess
import time
import requests
import logging
from . import util

l = logging.getLogger("pandem-admin")

def reset_variables():
  dir_path = util.pandem_path("files", "variables")
  home_path = util.pandem_path("files", "variables", "variables.csv")
  if not os.path.exists(dir_path):
    os.makedirs(dir_path)
  pkg_path = pkg_resources.resource_filename("pandemsource", "data/list-of-variables.csv")
  shutil.copy(pkg_path, home_path)
  # rewriting default variables
  reset_default_folders("variables", "indicators", delete_existing = False)

def reset_default_folders(*folders, delete_existing = True):
  # copy default data from data folder
  for folder in folders:
    if pkg_resources.resource_exists("pandemsource", f"data/{folder}"):
      var_from = pkg_resources.resource_filename("pandemsource", os.path.join("data", folder))
      var_to = util.pandem_path("files", folder)
      if os.path.exists(var_to) and delete_existing:
        shutil.rmtree(var_to)
      shutil.copytree(var_from, var_to, copy_function = shutil.copy, dirs_exist_ok = True)

def read_variables_definitions():
  path = util.pandem_path("files", "variables", "variables.csv")
  df = pd.read_csv(path, encoding = "ISO-8859-1")
  df = df.rename(columns = {
    "Variable":"variable", 
    "Data Family":"data_family", 
    "Linked Attributes":"linked_attributes", 
    "Partition":"partition", 
    "Aliases":"aliases", 
    "Description":"description", 
    "Type":"type", 
    "Unit":"unit", 
    "Datasets":"datasets"
    })

  for col in df.columns:
    if col not in ["description", "modifiers", "formula"] and not df[col].isnull().values.all():
      df[col] = df[col].str.lower().str.replace(", ", ",", regex=False).str.replace(".", "", regex=False).str.replace(" ", "_",regex=False)
    if col in ["linked_attributes", "partition"]:
      df[col] = df[col].str.split(",")
    if col == "modifiers":
      df[col] = df[col].apply(lambda x : json.loads(x) if pd.notna(x) else [])
  result = df.to_json(orient = "records")
  parsed = json.loads(result)
  return parsed

def reset_source(source_name):
  dls_from = pkg_resources.resource_filename("pandemsource", os.path.join("data", "DLS", f"{source_name}.json"))
  if os.path.exists(dls_from):
    dls_to = util.pandem_path("files", "source-definitions", f"{source_name}.json")
    dls_to_dir = util.pandem_path("files", "source-definitions")
    if not os.path.exists(dls_to_dir):
      os.makedirs(dls_to_dir)
    shutil.copyfile(dls_from, dls_to)
  else:
    raise ValueError(f"Cannot find source definition {source_name} within pandem default sources")
  
  # copytin script if any
  with open(util.pandem_path("files", "source-definitions", f"{source_name}.json"), "r") as f:
    dls = json.load(f)
    
  if "changed_by" in dls["acquisition"]["channel"] and  "script_type" in dls["acquisition"]["channel"]["changed_by"]:
    script_type = dls["acquisition"]["channel"]["changed_by"]["script_type"]
    script_name = dls["acquisition"]["channel"]["changed_by"]["script_name"]
    script_from = pkg_resources.resource_filename("pandemsource", os.path.join("data", "scripts", script_type, f"{script_name}.{script_type}"))
    script_to = util.pandem_path("files", "scripts", script_type, f"{script_name}.{script_type}" )
    script_to_dir = util.pandem_path("files", "scripts", script_type)
    if not os.path.exists(script_to_dir):
      os.makedirs(script_to_dir)
    shutil.copyfile(script_from, script_to)

  # Copy the corresponding Python script from package to pandem-home
  pyscript_fol = pkg_resources.resource_filename("pandemsource", os.path.join("data", "scripts", "py", "sources"))
  dls_pyscript = dls["scope"]["source"].replace("-", "_").replace(" ", "_") + ".py"
  pyscript_to = util.pandem_path("files", "scripts", "py", "sources", dls_pyscript)
  if dls_pyscript in os.listdir(pyscript_fol):
    shutil.copyfile(os.path.join(pyscript_fol, dls_pyscript), pyscript_to)


def delete_all():
    if os.path.exists(util.pandem_path("settings.yml")):
      os.remove(util.pandem_path("settings.yml"))
    if os.path.exists(util.pandem_path("files")):
      shutil.rmtree(util.pandem_path("files"))
    if os.path.exists(util.pandem_path("database")):
      shutil.rmtree(util.pandem_path("database"))

def install_issues(check_nlp = True):
  settings = util.settings()
  ret = []
  FNULL = open(os.devnull, 'w') 
  if shutil.which("R") is None:
    ret.append("Cannot find R language. Please installe it. PANDEM2 needs it to calculate indecators")
  else:
    r_packages = '"dplyr", "shiny", "plotly", "DT", "jsonlite", "httr", "XML", "ggplot2", "epitweetr"'
    installed = subprocess.run(['R', '-e', f'if(length(setdiff(c({r_packages}), names(installed.packages()[,1])))> 0) stop("some packages are missing!!")'], stdout=FNULL, stderr=FNULL).returncode
    if installed == 1:
      ret.append(f'Cannot find some necessary R packages, please intall them from CRAN, by running install.packages(c({r_packages}))')
    installed = subprocess.run(['R', '-e', f'if(length(setdiff(c("COVID19"), names(installed.packages()[,1])))> 0) stop("some packages are missing!!")'], stdout=FNULL, stderr=FNULL).returncode
    if installed == 1:
      ret.append(f"""Cannot find COVID19 R packages necessary for getting COVID19 data hub data. 
      if you want to download data directly from sources you have to install it as follow:
        install.packages("devtools")
        devtools::install_github(repo = "covid19datahub/COVID19", ref = "b941b66e59b7b0aec4807eb5b28208abba66de4a", upgrade = "never")
      If you prefer downloading data prepared by COVID data hub tem you can install it from CRAN, by running install.packages(c("COVID19"))""")
    installed = subprocess.run(['R', '-e', f'if(length(setdiff(c("Pandem2simulator"), names(installed.packages()[,1])))> 0) stop("some packages are missing!!")'], stdout=FNULL, stderr=FNULL).returncode
    if installed == 1:
      ret.append(f"""Cannot find Pandem2simulator R packages necessary for simulating COVID19 detailed data from ecdc datasets. 
      You can install this package with the following command:
        install.packages("devtools")
        
      devtools::install_github("maous1/Pandem2simulator")
      """)
  if settings["pandem"]["source"]["nlp"]["active"] and check_nlp:
     models_path = settings["pandem"]["source"]["nlp"]["models_path"]
     if os.path.exists(models_path):
       if not nlp_models_up():
         # Here if it will be to pandem2 to launch docker since models are not up   
         if shutil.which("docker") is None:
           ret.append("Cannot find docker and NLP models are not running, we need it to start the NLP server. Please install it.")
         else:
           l.info("Launching docker for serving NLP modles")
           retries = 10
           threading.Thread(target=nlp_run_model_server).start()
           while retries > 0 and not nlp_models_up():
             l.info("Waiting 1 until models are running")
             time.sleep(1)
             retries = retries - 1
       if not nlp_models_up():
         ret.append(f"""Launching of docker NLP server failed. Please check any previous errors.
           If you cannot (or do not want to) fix this, you can deactivate the NLP models on the settings.json file by setting the property by changing pandem.source.nlp.active to False 
           un {util.pandem_path('settings.yml')}
           To fix this ensure that:
             - Docker is installed https://docs.docker.com/engine/install/
             - You have downloaded the PANDEM-2 sma components to a local folder https://drive.google.com/file/d/1mSl2X4DQQZKf1sHeJaKDZOM6ydi4nVEK/view?usp=sharing
             - You have either set the PANDEM_NLP environment variable to the folder holding the models or set the value pandem.source.nlp.models.path on  {util.pandem_path('settings.yml')}
             - You can test the launching command as follow
               {nlp_docker_launch_command()}
          """
         )
     else:
       ret.append("""NLP Annotation is active as per settings but the models path has not been found which is necessary to detect the existing models (even when running the server outside PANDEM-2
             You have either set the PANDEM_NLP environment variable to the folder holding the models or set the value pandem.source.nlp.models.path on  {util.pandem_path('settings.yml')}
       """)
  if check_nlp and are_twitter_credentials_missing():
    ret.append(f"""Twitter credentials are necessary since one of your sources uses twitter, but we could not find the credentials
      please try running PANDEM-2 again to provide the right credentials or remove the twitter data source definition on {util.pandem_path("files", "source-definitions")}
      To get twitter credentials please check you have to decralre it. Please check this: https://developer.twitter.com/en/docs/twitter-api/getting-started/getting-access-to-the-twitter-api
    """)

  return ret

def nlp_models_up():
  settings = util.settings()
  models_path = settings["pandem"]["source"]["nlp"]["models_path"]
  models = list(filter(lambda v: not v.startswith("."), next(os.walk(models_path))[1]))
  tf_url = f"{settings['pandem']['source']['nlp']['tensorflow_server_protocol']}://{settings['pandem']['source']['nlp']['tensorflow_server_host']}:{settings['pandem']['source']['nlp']['tensorflow_server_port']}"
  model_endpoints = {m:f"{tf_url}/v1/models/{m}" for m in models}

  for ep in model_endpoints.values():
    try:
      if requests.get(ep).status_code != 200:
        return False
    except Exception as err:
      return False
  return True

def nlp_docker_launch_command():
  settings = util.settings()
  tf_usesudo = settings["pandem"]["source"]["nlp"]["use_sudo"]
  tf_port = settings["pandem"]["source"]["nlp"]["tensorflow_server_port"]
  tf_version = settings["pandem"]["source"]['nlp']["tensorflow_server_version"]
  models_path = settings["pandem"]["source"]["nlp"]["models_path"]
  models = list(filter(lambda v: not v.startswith("."), next(os.walk(models_path))[1]))
  
  if os.name == "posix" and tf_usesudo:
    cmd = ["sudo"]
  else: 
    cmd = []
  cmd.extend(["docker", "run", "-p", f"{tf_port}:8501"]) 
  
  for model in models:
    cmd.extend(["-v",  f"{models_path}{os.sep}{model}:/models/{model}/1/"]) 

  cmd.extend(["-v", f"{models_path}{os.sep}models.config:/models/models.config"])
  cmd.extend(["--rm", "-t", f"tensorflow/serving:{tf_version}", "--model_config_file=/models/models.config"])
  return cmd

def nlp_run_model_server():
  #sudo_password = getpass.getpass(prompt='enter your sudo password: ')
  #print("pwd received!")
  cmd = nlp_docker_launch_command()
  l.debug(f"Launching docker comman for nlp model server: {cmd}")
  #p = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE,  stdin=subprocess.PIPE)
  #with open(os.devnull, 'w') as devnull:
  p = subprocess.run(cmd)
  
  #try:
  #  out, err = p.communicate(input=(sudo_password+'\n').encode(),timeout=20)
  #except subprocess.TimeoutExpired:
  #  print("\n\n\n\n\n\n BUAAAAAAAAAAAAAAAAAAAAAA")
  #  p.kill()


def are_twitter_credentials_missing():
   dls_path = util.pandem_path("files", "source-definitions")
   if len([
       p for 
       p in os.listdir(dls_path) 
       if p.endswith(".json") 
         and json.load(open(os.path.join(dls_path, p)))["acquisition"]["channel"]["name"] == "twitter"]
   ) > 0:
     l.info("Checking if twitter API are necessary")
     secrets = [
       util.get_or_set_secret("twitter-api-key"), 
       util.get_or_set_secret("twitter-api-key-secret"), 
       util.get_or_set_secret("twitter-access-token"),
       util.get_or_set_secret("twitter-access-token-secret")
     ]
     return any(filter(lambda s: s == "", secrets))
   return False

def run_pandem2app():
  settings = util.settings()
  cmd = ["Rscript", util.pandem_path("files", "scripts", "R", "pandem-source-app.R"), str(settings["pandem"]["source"]["app"]["port"]), util.pandem_path("files", "img", "Partner-map_PANDEM-2.jpg")]
  l.debug(f"Launching app server: {cmd}")
  def run_app(): 
    subprocess.run(cmd)
  threading.Thread(target=run_app).start()


def list_sources_dir(p):
  ret = dict()
  for f in os.listdir(p):
   if f.endswith(".json"):
     with open(os.path.join(p, f)) as fj:
       js = json.load(fj)
       if "scope" in js:
         s, t = (js["scope"].get("source"), js["scope"].get("tags"))
         ret[s] = t[0] if len(t) > 0 else s
  return ret

def list_sources(local = True, default = False, missing_local = False, missing_default = False):
  local_map = list_sources_dir(util.pandem_path("files", "source-definitions"))
  default_map = list_sources_dir(pkg_resources.resource_filename("pandemsource", os.path.join("data", "DLS")))
  
  if local and missing_local or default and missing_default:
    raise ValueError("Cannot lis both missing and existing sources")
  ret = list()
  if local:
    ret.extend(local_map.items())
  elif missing_local:
    ret.extend((k, v) for k, v in default_map.items() if k not in local_map)
  if default:
    ret.extend(local_map.items())
  elif missing_default:
    ret.extend((k, v) for k, v in local_map.items() if k not in defaul_map)
  return ret






import os
from . import worker
from abc import ABC, abstractmethod, ABCMeta
from datetime import datetime, timedelta
import time
import logging as l

class Acquisition(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings, channel): 
        self.channel = channel
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)    
         
    def on_start(self):
        super().on_start()
        self.current_sources = dict()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy() 

    
    @abstractmethod
    def new_files(self, dls, last_hash):
        pass

    def source_path(self, dls, *args):
        if dls['acquisition']['channel']['name']=='input-local':
            return self.pandem_path(f'files/{self.channel}', *args)
        else:
            return self.pandem_path(f'files/{self.channel}', dls['scope']['source'], *args)

    def monitor_source(self, source_id, dls, freq): 
        #print(f'here acquisition monitor_source loop: {source_id} {freq}')
        last_hash = self._storage_proxy.read_db('source', lambda x: x['id']==source_id).get()['last_hash'].values[0]
        #Getting new files if any
        nf = self.new_files(dls, last_hash)
        files_to_pipeline = nf["files"]
        #print(f'files to pipeline: {files_to_pipeline}')
        new_hash = nf["hash"]
        # If new files are found they will be send to the pipeline 
        if len(files_to_pipeline)>0:
            l.info(f'New {len(files_to_pipeline)} hash changed from {last_hash} to {new_hash}')
            # Sending files to the pipeline
            self._pipeline_proxy.submit_files(dls, files_to_pipeline).get()
            # Storing the new hash into the db
            self._storage_proxy.write_db(
                {'name': dls['scope']['source'],
                    'last_hash': new_hash,
                    'id': source_id
                }, 
                'source'
            ).get()
        self._storage_proxy.write_db(
           {'id': source_id,
               'last_exec': freq.last_exec,
               'next_exec': freq.next_exec()
           }, 
           'source'
        ).get()  


    def add_datasource(self, dls, force_acquire):
        source_dir = self.source_path(dls)
        if not os.path.exists(source_dir):
            os.makedirs(source_dir)
        # registering new source if not already on the source table
        find_source = self._storage_proxy.read_db('source', lambda x: x['name']==dls['scope']['source']).get()
        if find_source is None  or len(find_source["id"]) == 0:
            last_exec = None
            id_source = self._storage_proxy.write_db(
                 {
                  'name': dls['scope']['source'],
                  'last_hash':""
                 },
                'source'
            ).get()
        else:
            if force_acquire:
                for s in find_source.to_dict(orient='records'):
                  s["last_hash"] = ""
                  s["last_exec"] = None
                  s["next_exec"] = None
                  self._storage_proxy.write_db(s, "source").get()

            last_exec = find_source["last_exec"].values[0]
            id_source = find_source["id"].values[0]

       # reseting the last_hash if force acquire is requested
       # updating the internal variable current sources
        self.current_sources[id_source] = dls 
        
        loop_frequency_str = dls['scope']['frequency'].strip()
        if 'frequency_start_hour' in dls['scope']:
          start_hour = int(dls['scope']['frequency_start_hour'])
        else :
          start_hour = None
        if 'frequency_end_hour' in dls['scope']:
          end_hour = int(dls['scope']['frequency_end_hour'])
        else :
          end_hour = None
        if loop_frequency_str=='':
            monitor_repeat = worker.Repeat(timedelta(days=1), start_hour = start_hour, end_hour = end_hour, last_exec = last_exec)         
        elif loop_frequency_str=='daily':
            monitor_repeat = worker.Repeat(timedelta(days=1), start_hour = start_hour, end_hour = end_hour, last_exec = last_exec)
        elif loop_frequency_str.split(' ')[-1]=='seconds':
            loop_frequency = int(loop_frequency_str.split(' ')[-2]) #every n seconds
            monitor_repeat = worker.Repeat(timedelta(seconds=loop_frequency), start_hour = start_hour, end_hour = end_hour, last_exec = last_exec)
        elif loop_frequency_str.split(' ')[-1]=='minutes':
            loop_frequency = int(loop_frequency_str.split(' ')[-2]) #every n minutes
            monitor_repeat = worker.Repeat(timedelta(minutes=loop_frequency), start_hour = start_hour, end_hour = end_hour, last_exec = last_exec)
        elif loop_frequency_str.split(' ')[-1]=='hours':
            loop_frequency = int(loop_frequency_str.split(' ')[-2]) #every n hours
            monitor_repeat = worker.Repeat(timedelta(hours=loop_frequency), start_hour = start_hour, end_hour = end_hour, last_exec = last_exec)

        self.register_action(repeat=monitor_repeat, action=lambda: self.monitor_source(id_source, dls, monitor_repeat))        






from . import worker
import os
from . import util 
from .storage import CacheValue
from . import admin
import itertools
import json
import datetime
import numpy
from collections import defaultdict
from .util import JsonEncoder
import logging
import pickle
import copy

l = logging.getLogger("pandem.variable")

class Variables(worker.Worker):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)

    def on_start(self):
        super().on_start()
        self._storage_proxy=self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._pipeline_proxy=self._orchestrator_proxy.get_actor('pipeline').get().proxy()
        self._variables = None
        self._timeseries = None
        self._timeseries_outdated = False
        self.timeseries_hash = ''

    def get_variables(self):
        if self._variables is None:
          dic_variables = dict()
          var_list=admin.read_variables_definitions()
          for var in var_list: 
              if not var["base_variable"]:            
                  base_dict = var.copy()
                  base_dict["aliases"] = []
                  base_dict.pop("base_variable")
                  aliases = [{"alias":v['variable'], 
                              "variable": v['base_variable'],
                              "formula": v['formula'],
                              "modifiers": v['modifiers'],
                              "description": v['description'],
                              "type": v['type']
                              }
                              for v in var_list if v['base_variable']==var['variable']]
                  
                  base_dict["aliases"] = aliases
                  dic_variables[var['variable']] = base_dict
                  if aliases:
                      for alias in aliases:
                          alias_dict = base_dict.copy()
                          alias_dict['formula'] = alias['formula'] if alias['formula'] is not None else None #base_dict['formula']
                          alias_dict['modifiers'] = alias['modifiers']
                          alias_dict['description'] = alias['description'] if alias['description'] is not None else alias_dict['description']
                          alias_dict['type'] = alias['type'] if alias['type'] is not None else alias_dict['type']
                          dic_variables[alias['alias']] = alias_dict
          self._variables = dic_variables
        return self._variables

    def get_referential(self,variable_name):
        #print(f"getting {variable_name}")
        list_files=[]
        referentiel=[]
        path=os.path.join(os.getenv('PANDEM_HOME'), 'files/variables/', variable_name)
        if os.path.isdir(path):
            list_files=self._storage_proxy.list_files(path).get()
            for file in list_files:
                var_list=self._storage_proxy.read_file(file['path']).get()
                if type(var_list) == dict:
                    for var in var_list['tuples']:
                        referentiel.append(var)
        else: 
            return None
        return referentiel


    def read_variable(self,variable_name, filter = {}):
        # l.debug(f"requesting {variable_name} with filter = {filter}" )       
        dir_path = util.pandem_path('files/variables/', variable_name)
        variables = self.get_variables()
        if os.path.isdir(dir_path):
            for mod in variables[variable_name]["modifiers"]:
              if mod['variable'] in filter:
                filter = filter.copy()
                filter[mod['variable']] = {mod['value']}
            requested_list = []
            list_files = self._storage_proxy.list_files_abs(dir_path).get()
            target_files = []
            for file in list_files:
                file_name = file['name']
                if file_name == 'default.json':
                    target_files.append(file)
                elif file_name.endswith(".json"):
                    can_ignore = False
                    for var_val in file_name.split('.')[:-1]:
                        var = var_val.split('=')[0]
                        val = var_val.split('=')[1]
                        if var in filter and type(filter[var]) in [list, set] and val not in filter[var]:
                            can_ignore = True
                            break
                        elif var in filter and type(filter[var]) not in [list, set] and filter[var] != val:
                            can_ignore = True
                            break
                    if not can_ignore:
                        target_files.append(file)
            for file in target_files:
                var_tuples = self._storage_proxy.read_file(file['path']).get()['tuples']
                requested_vars = []
                for var_tuple in var_tuples:
                    if all(att in var_tuple['attrs'].keys() for att in filter.keys()):
                        if all((val is None or var_tuple['attrs'][att]==val or var_tuple['attrs'][att] in val) for att, val in filter.items()) :
                            requested_vars.append(var_tuple)
                requested_list.extend(requested_vars)
            return requested_list
        else: 
            return None

    

    
    def get_partition(self, tuple, partition):
        if partition is None:
          return 'default.json'
        else:
          return '.'.join([key + '=' + str(val) for key, val in tuple['attrs'].items() if key in partition]) + '.json'

    def remove_attrs(self, t, private_attrs):
        if "attrs" in t:
          if t["attrs"].keys().isdisjoint(private_attrs):
            return t
          else:
            tt = t.copy()
            tt["attrs"] = {k:v for k,v in t["attrs"].items() if k not in private_attrs}
            return tt

    def apply_aliases(self, t, aliases):
        applied = False
        if "obs" not in t or t["obs"].keys().isdisjoint(aliases.keys()):
          return t
        else:
          for var, value in t["obs"].items():
              tt = copy.deepcopy(t)
              if "attrs" not in tt:
                 tt["attrs"] = {}
              # renaming the observation to its base variable
              if var != aliases[var]["variable"]:
                tt["obs"][aliases[var]["variable"]] = tt["obs"].pop(var)
              
              # applyting modifiers
              for modifier in aliases[var]["modifiers"]:
                tt["attrs"][modifier['variable']] = modifier['value']
          return tt
 
    def write_variable(self, input_tuples, step, job):
        # if input_tuples is in cache get its value
        if type(input_tuples) == CacheValue:
          input_tuples = input_tuples.value()

        variables = self.get_variables()
        partition_dict = defaultdict(list)
        if step ==0:
            self.write_variable(self.tag_source_var(job["dls_json"]), None, job)

        if input_tuples['tuples']:
            # building a dictionary with destination files for tuples depending on variable name at their partitioning schema
            private_attrs = {k for k, v in variables.items() if v["type"] == "private"}
            aliases = {v:vardic for v, vardic in variables.items() if "modifiers" in vardic and len(vardic["modifiers"]) > 0}
            for t in input_tuples['tuples']:
                # removing private attributes
                t = self.remove_attrs(t, private_attrs)
                # applying modifiers if tuple pbservation contains modifiers
                t = self.apply_aliases(t, aliases)
                # associating this tuple tuple to its destination variable
                var_name = None
                for key, var in t.items():
                    if key != "attrs" and len(t[key].keys()) > 0:
                        var_name = list(t[key].keys())[0]
                if var_name is not None:
                  partition_dict[var_name].append(t)

            # building the dict tuples per destination files
            partition_dict_final = defaultdict(lambda: defaultdict(list))
            for var, tuple_list in partition_dict.items():
                for t in tuple_list:
                    file_name = self.get_partition(t, variables[var]["partition"])
                    partition_dict_final[var][file_name].append(t)
            
            # identifying the scope of data that will be replaced by current tuples
            update_filter = []
            for filter in input_tuples['scope']['update_scope']:
                if not isinstance(filter['value'], list):
                    update_filter.append({'variable':filter['variable'], 'value':[str(filter['value'])]})
                else:
                    update_filter.append({'variable':filter["variable"], 'value':list([str(f) for f in filter["value"]])})
            
            # Iterating on each variable to write
            for var, tuples_dict in partition_dict_final.items():
                var_dir = util.pandem_path('files/variables', var)
                if not os.path.exists(var_dir):
                    os.makedirs(var_dir)
                # Iterating on each file to write
                for file_name, tuples_list in tuples_dict.items():
                    if 'attr' in tuples_list[0]:
                        # this is for avoiding having duplicates on referentials (attr instead of obs)
                        # TODO: Improve this using the variable type
                        unique_values = {}
                        for index, t in enumerate(tuples_list):
                            if t['attr'][var] not in unique_values.values():
                                unique_values[index] = t['attr'][var]
                        tuples_list = [tuples_list[key] for key in unique_values.keys()]

                   
                    file_path = util.pandem_path(var_dir, file_name)
                    # If file does not exists then we can just dump the tuples (no need to delete) 
                    tuples_to_dump = {'tuples': tuples_list}
                    if not os.path.exists(file_path) or os.path.getsize(file_path) == 0:
                        with open(file_path, 'w+') as f:
                            json.dump(tuples_to_dump, f, cls=JsonEncoder, indent = 4)
                    # If file exists already then we need to remove lines on the replacement scope
                    # We do that by appending to tuples_list any existing row not in the replacement scope and then overriding the file 
                    else:
                        with open(file_path, 'r') as f:
                            last_tuples = json.load(f)
                        for tup in last_tuples['tuples']:
                            cond_count = len(update_filter)
                            for filt in update_filter: 
                                if filt['variable'] in tup['attrs'].keys() and tup['attrs'][filt['variable']] in filt['value']:
                                    cond_count = cond_count - 1
                            # addding the tuple of one of the conditions failed
                            if cond_count > 0:
                                tuples_list.append(tup)            
                        # replacing the file
                        with open(file_path, 'w') as f:
                            json.dump(tuples_to_dump, f, cls=JsonEncoder, indent = 4)
                   
                    # TODO: get deleted files and remove from time series before adding new ones
                    # Now that we have updated the file we can update the time series cache
                    self.get_timeseries(tuples=tuples_to_dump['tuples'], tuples_path=file_path, save_changes = False)

        # saving changes done on the time series cache
        self.get_timeseries(save_changes = True)
        if step is not None:
            self._pipeline_proxy.publish_end(job = job)
                    
    def tag_source_var(self, dls):
       tags = dls['scope']['tags'] if 'tags' in dls['scope'] else []
       source = dls['scope']['source']
       return {
         'scope':{
           'update_scope':[
             {'variable':'source', 'value':source}
           ]
         },
         'tuples':[*({'attr':{'tag_source': tag+'_'+source}, 'attrs':{'tag':tag, 'source': source}} for tag in tags)]
       }

    def lookup(self, variables, combinations, source, filter = None, include_source = True, include_tag = False, types=None):
      dico_vars  = self.get_variables()
      modifiers = {v:{m['variable']:m['value'] for m in vardef['modifiers']} for v, vardef in dico_vars.items() if len(dico_vars[v]["modifiers"]) > 0}
      tag_source = self.read_variable("tag_source")
      tags = {t['attrs']['tag'] for t in tag_source if t['attrs']['source'] == source}
      others = {t['attrs']['source'] for t in tag_source if t['attrs']['tag'] in tags and t['attrs']['source'] != source}

      attrs = set()
      filt = {}
      if "attrs" in combinations:
        indexed_comb = {}
        for comb in combinations:
          if "attrs" in comb:
            keys = sorted(comb["attrs"].keys())
            attrs.update(keys)
            key = tuple((k, comb[k]) for k in keys if types is None or dico_vars[k]["type"] in types)
            for k, v in key.items():
              if k not in filt:
                filt[k] = {}
              filt[k].add(v)
            indexed_comb[key] = comb["obs"] if "obs" in comb else {}
      else:
        indexed_comb = combinations
        for key in indexed_comb:
          for k, v in key:
            if k not in filt:
              filt[k] = set()
            filt[k].add(v)
          if types is None:
            attrs.update(k for k, v in key)

      if types is None:
        types = {dico_vars[t]["type"] for t in attrs}
      if filter is not None:
        filt.update(filter)
      res = {}
      
      for var in variables:
        tuples = []
        if include_source:
          f = {k:v for k, v in filt.items() if v is not None}
          f.update({"source":source})
          # TODO make read variable work with alias variables so it can return active cases and confirmed cases on the same query
          tuples = self.read_variable(var, filter = f)
        if include_tag and (tuples is None or len(tuples) == 0):
          f = {k:v for k, v in filt.items() if v is not None}
          f.update({"source":others})
          tuples = self.read_variable(var, filter = f)
        # if currnet variable has modifiers we have to find the matching index for the tuple
        key_map = {tuple((k, (v if var not in modifiers or not k in modifiers[var] else modifiers[var][k])) for k, v in comb):comb for comb in combinations}
        if tuples is not None:
          for t in tuples:
            keys = sorted(attr for attr in t["attrs"].keys() if dico_vars[attr]["type"] in types and (t["attrs"][attr] is not None or (var in modifiers and attr in modifiers[var])))
            if tuple((k, t["attrs"][k]) for k in keys) in key_map:
              key = key_map[tuple((k, t["attrs"][k]) for k in keys)]
              filter_value = {k:v for k, v in t["attrs"].items() if k in (filter if filter is not None else [])}  
              if key in indexed_comb:
                if not key in res:
                  res[key] = {}
                if "obs" in t:
                  obs_name =  next(iter(t["obs"].keys()))
                  if not obs_name in res[key]:
                    res[key][obs_name] = []
                  res[key][obs_name].append({"value":t["obs"][obs_name], "attrs":filter_value})
      return res
      

    def get_timeseries(self, tuples=None, tuples_path=None, save_changes = True):
      var_dic = self.get_variables()
      storage_proxy  = self._storage_proxy

      if tuples is not None and tuples_path is None:
        raise ValueError("If tuples is None, tuple path is needed in order to register the path new tuples")
      if tuples is None and tuples_path is not None:
        tuples = storage_proxy.read_file(util.pandem_path(tuples_path)).get()['tuples']
      
      cache_path = util.pandem_path('files', 'variables', 'time_series.pi')
      
      # restoring current time series from memory or disk
      if self._timeseries is not None:
        cache = self._timeseries
      elif storage_proxy.exists(cache_path).get():
        with open(cache_path, 'rb') as cf:
          cache = pickle.load(cf)
        self._timeseries = cache
      else:
        # If there is no currently any cahe we have to build it
        # by reading all published variables 
        cache = {}
        self._timeseries_outdated = True
        self._timeseries = cache
        for var, varinfo in var_dic.items(): 
          if varinfo["type"] in ["observation", "indicator", "resource"] and varinfo["variable"] == var:
            var_path = util.pandem_path('files', 'variables', var)
            if storage_proxy.exists(var_path).get():
               l.debug(f'Producing timeseries cache for {var}')
               for finfo in storage_proxy.list_files(util.pandem_path('files', 'variables', var_path)).get():
                 t = storage_proxy.read_file(finfo["path"]).get()
                 if type(t) == dict:
                   self.get_timeseries(tuples = t['tuples'], tuples_path = finfo['path'], save_changes = False)
      
      # If no new tuples no analyze return cache 
      if tuples is not None:
          for t in tuples:
            if "obs" in t and len(t["obs"]) > 0 and "attrs" in t:
              var_name = next(iter(t["obs"]))
              if var_name in var_dic and "aliases" in var_dic[var_name]:
                modifiers = []
                obs_name = var_name
                
                # Fitting an alias if possible
                for alias in var_dic[var_name]["aliases"]:
                  if len(alias['modifiers']) > len(modifiers):
                    if all(m['variable'] in t['attrs'] and (m['value'] is None or t['attrs'][m['variable']] == m['value']) for m in alias['modifiers']):
                      obs_name = alias['alias']
          
                # getting tuple key
                key = []
                key.append(("indicator", obs_name))
                to_ignore = ['not_characteristic', 'date']
                key.extend([(k, v) for k,v in t["attrs"].items() if v is not None and k in var_dic and var_dic[k]["type"] not in to_ignore or k == "source"])
                key.sort(key = lambda p:p[0])
                key = tuple(key)
          
          
                # getting dates
                dates = [str(v) for k,v in t["attrs"].items() if k in var_dic and var_dic[k]["type"] == "date"]
                if len(dates)>0:
                  min_date = min(dates)
                  max_date = max(dates)
                else:
                  min_date = None
                  max_date = None
                
                tuples_relpath = os.path.relpath(tuples_path, util.pandem_path())
                # storing the series key on the cache with its statistics (min date, max date and files)
                if key not in cache:
                  self._timeseries_outdated = True
                  cache[key] = {"paths":{tuples_relpath}, "min_date":min_date, "max_date":max_date}
                else:
                  if tuples_relpath not in cache[key]["paths"]:
                    cache[key]["paths"] = cache[key]["paths"].union({tuples_relpath})
                    self._timeseries_outdated = True
                  if min_date is not None and min_date != cache[key]["min_date"]:
                    cache[key]["min_date"] = min(min_date, cache[key]["min_date"]) if cache[key]["min_date"] is not None else min_date
                    self._timeseries_outdated = True
                  if max_date is not None and max_date != cache[key]["max_date"]:
                    cache[key]["max_date"] = max(max_date, cache[key]["max_date"]) if cache[key]["max_date"] is not None else max_date
                    self._timeseries_outdated = True

      # Saving changes to picke if any change found and save_changes is requested
      if self._timeseries_outdated and save_changes:
        with open(cache_path, 'wb') as f:
          pickle.dump(cache, f)
        self._timeseries_outdated = False
        self.timeseries_hash = str(datetime.datetime.now())
      return cache  



import os
from . import worker
from abc import ABC, abstractmethod, ABCMeta
import time
from zipfile import ZipFile
from io import BytesIO

class Unarchive(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
         
    def on_start(self):
        super().on_start()
        self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy()

    def loop_actions(self):
        pass
    
    def unarchive(self, archive_path, filter_path, job): 
        #print('here in unarchive')
        # print(f'files_to_filter: {files_to_filter}')
        with ZipFile(archive_path) as zip_archive:
            self._pipeline_proxy.unarchive_end(archive_path, filter_path, BytesIO(zip_archive.read(filter_path)).read(), job)


       





import numpy as np
import pandas as pd
import datetime

def df_transform(df):
  data = list()
  starts = df["date_start"]
  starts = starts.replace(to_replace="NA", value=pd.NA)
  starts = pd.to_datetime(starts)
  
  ends = df["date_end"]
  ends = ends.replace(to_replace="NA", value=pd.NA)
  ends = pd.to_datetime(ends)

  last = max(ends)
  for i in df.index:
    start = starts[i]
    end = ends[i] if pd.notna(ends[i]) else last
    dates = [end - datetime.timedelta(days = x) for x in range((end - start).days + 1)]
    dates.reverse()
    data.append((df["line_number"][i], df["Country"][i].upper(), df["Response_measure"][i], min(dates) - datetime.timedelta(days = 1), 0))
    for d in dates:
      data.append((df["line_number"][i], df["Country"][i].upper(), df["Response_measure"][i], d, 1))
    data.append((df["line_number"][i], df["Country"][i].upper(), df["Response_measure"][i], max(dates) + datetime.timedelta(days = 1), 0))
  newdf = pd.DataFrame(data, columns = ["line_number", "Country", "Response_measure", "date", "count"])
  return newdf




import pykka
import os
import time
import threading
import datetime
import json
import hashlib
from collections import defaultdict
from . import worker
from . import util
from abc import ABC, abstractmethod, ABCMeta

import logging

l = logging.getLogger("pandem.pipeline")

class Pipeline(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings, retry_failed = False, restart_job = 0, retry_active = True): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
        self.retry_failed = retry_failed
        self.retry_active = retry_active
        self.restart_job = restart_job
        self._jobs_to_keep = settings["pandem"]["source"]["pipeline"]["jobs-to-keep"]

    def on_start(self):
        super().on_start()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._frxml_proxy = self._orchestrator_proxy.get_actor('ftreader_xml').get().proxy()
        self._frcsv_proxy = self._orchestrator_proxy.get_actor('ftreader_csv').get().proxy()
        self._frxls_proxy = self._orchestrator_proxy.get_actor('ftreader_xls').get().proxy()
        self._frjson_proxy = self._orchestrator_proxy.get_actor('ftreader_json').get().proxy()
        self._unarchive_proxy = self._orchestrator_proxy.get_actor('unarchiver').get().proxy()
        self._dfreader_proxy = self._orchestrator_proxy.get_actor('dfreader').get().proxy()
        self._standardizer_proxy = self._orchestrator_proxy.get_actor('standardizer').get().proxy()
        self._variables_proxy = self._orchestrator_proxy.get_actor('variables').get().proxy()
        self._nlp_proxy = self._orchestrator_proxy.get_actor('nlp_annotator').get().proxy()
        self._aggregate_proxy = self._orchestrator_proxy.get_actor('aggregator').get().proxy()
        self._evaluator_proxy = self._orchestrator_proxy.get_actor('evaluator').get().proxy()

        self.job_steps = defaultdict(dict)
        self.decompressed_files = defaultdict(list)
        self.job_precalstep = {}
        self.job_df = defaultdict(dict)
        self.job_tuples = defaultdict(dict)
        self.job_stdtuples = defaultdict(dict)
        self.job_aggrtuples = {}
        self.job_precalinds = {}
        self.job_indicators = {}
        self.job_issues = defaultdict(list)
        self.pending_count = {}

        

        self.job_dicos = [
          self.decompressed_files, self.job_df, self.job_tuples, self.job_stdtuples, self.job_aggrtuples, 
          self.job_precalinds, self.job_indicators, self.pending_count, self.job_issues
        ]

        self.last_step = "ended"
        jobs = self._storage_proxy.read_db('job',
                                          filter= lambda x: 
                                             ( 
                                                 ((x['status'] == 'in progress' and self.retry_active) or (x['status'] == 'failed' and self.retry_failed)) 
                                                   and x['step'] != 'submitted_started' 
                                                   and x['step'] != self.last_step
                                                   and self.restart_job == 0
                                             ) or x["id"] == self.restart_job
                                         ).get()
        if jobs is not None :
            jobs = jobs.to_dict(orient = 'records')

            new_dls = dict(
              (dls["scope"]["source"],dls) 
              for dls in  (
                self._storage_proxy.read_file(f["path"]).get() for f in self._storage_proxy.list_files('source-definitions').get() if f['path'].endswith(".json"))
              )
            for j in jobs:
              j["step"]="submitted_ended"
              j["status"]="in progress"
              j["dls_json"] = new_dls[j["source"]]
              # deleting issues on jobs we are restarting
              self._storage_proxy.delete_db('issue', filter = lambda i: int(i['job_id']) == int(j["id"])) 

            self.job_steps['submitted_ended'] = dict([(j["id"], j) for j in jobs])  
        process_repeat = worker.Repeat(datetime.timedelta(seconds=1))
        self.register_action(process_repeat, self.process_jobs) 
    

    def loop_actions(self):
        self._self_proxy.process_jobs()

    def process_jobs(self):
        # This function will process active jobs asynchronoulsy.
        # Based on current status an action will be performed and the status will be updated
        # Jobs after submit that has just been submitted 
        # they will go to decompress or to format read
        for job in self.job_steps['submitted_ended'].copy().values():
            submitted_files = job['source_files']
            if "decompress" not in job['dls_json']['acquisition'].keys():
                self.update_job_step(job, 'read_format_started', 0.1)
                self.send_to_readformat(submitted_files, job)
            else:
                self.update_job_step(job, 'unarchive_started', 0.1)
                self.send_to_unarchive(submitted_files, job)
        
        # Jobs after decompression ends
        for job in self.job_steps['unarchive_ended'].copy().values():
            decompressed_files = self.decompressed_files[job['id']]
            self.update_job_step(job, 'read_format_started', 0.2)
            self.send_to_readformat(decompressed_files, job)

        # Jobs after read format ends
        for job in self.job_steps['read_format_ended'].copy().values():
            dfs = self.job_df[job["id"]]
            self.update_job_step(job, 'read_df_started', 0.3)
            self.send_to_read_df(dfs, job)
        
        # Jobs after read df ends
        for job in self.job_steps['read_df_ended'].copy().values():
            tuples = self.job_tuples[job["id"]]
            self.update_job_step(job, 'standardize_started', 0.4)
            self.send_to_standardize(tuples, job)
        
        # Jobs after standardize ends
        for job in self.job_steps['standardize_ended'].copy().values():
            tuples = self.job_stdtuples[job["id"]]
            # non standardized tuples can be deleted
            self.job_tuples.pop(job["id"])
            self.job_df.pop(job["id"])
            if self.job_to_annotate(job):
                self.update_job_step(job, 'annotate_started', 0.5)
                self.send_to_annotate(tuples, job)
            else:
                self.update_job_step(job, 'aggregate_started', 0.6)
                self.send_to_aggregate(tuples, job)

        # Jobs after annnotate text ends
        for job in self.job_steps['annotate_ended'].copy().values():
            tuples = self.job_stdtuples[job["id"]]
            self.update_job_step(job, 'aggregate_started', 0.6)
            self.send_to_aggregate(tuples, job)
        
        # Jobs after aggregate ends
        for job in self.job_steps['aggregate_ended'].copy().values():
            # standardized tuples can be deleted
            self.job_stdtuples.pop(job["id"])
            tuples = self.job_aggrtuples[job["id"]]
            self.update_job_step(job, 'precalculate_started', 0.7)
            self.send_to_precalculate(tuples, job)
        
        # Jobs after precalculate ends
        for job in self.job_steps['precalculate_ended'].copy().values():
            tuples = self.job_aggrtuples[job["id"]]
            # aggregated tuples are not necessary anymore
            self.job_aggrtuples.pop(job["id"])

            self.update_job_step(job, 'publish_started', 0.9)
            self.send_to_publish(tuples, job)

        # Jobs after calculate ends
        for job in self.job_steps['calculate_ended'].copy().values():
            indicators_tuples = self.job_indicators[job["id"]]
            self.update_job_step(job, 'publish_started', 0.9)
            self.send_to_publish(indicators_tuples, job)

        # Jobs after publish indicator ended
        for job in self.job_steps['publish_ended'].copy().values():
            calc_step = self.job_precalstep[job['id']] + 1
            self.job_precalstep[job['id']] = calc_step
            # recreating the indicators to calculate for the current step
            ind_in_step = {k:v for k,v in self.job_precalinds[job['id']].items() if v["step"] == calc_step}
            if len(ind_in_step) > 0:
              self.update_job_step(job, 'calculate_started', 0.8)
              self.send_to_calculate(ind_in_step, job)
            else:       
              self.update_job_step(job, 'ended', 1.0)
              self.remove_job(job)
     
    def submit_files(self, dls, paths):
        job_record = {'source': dls['scope']['source'],
                      'file_sizes': [os.path.getsize(path)/1024 for path in paths],
                      'progress': 0.0,
                      'start_on': datetime.datetime.now(),
                      'end_on': None,
                      'step': 'submitted_started',
                      'status': 'in progress',
                      'dls_json' : dls
                     }
        job_id = self._storage_proxy.write_db(job_record, 'job').get()
        dest_files = [str(job_id)+'_'+'_'.join(path.split('files/')[1].split('/')) for path in paths]
        
        paths_in_staging = [self.staging_path(job_id, 'in', file) for file in dest_files]
        self._storage_proxy.copy_files(paths, paths_in_staging)
        job_id = self._storage_proxy.write_db({'id': job_id,
                                              'source_files': paths_in_staging,
                                              'step': 'submitted_ended'}, 
                                             'job').get()
        job = self._storage_proxy.read_db('job', filter= lambda x: x['id']==job_id).get().to_dict(orient='records')[0]

        self.update_job_step(job, 'submitted_ended', 0.05)

    def send_to_readformat(self, paths, job):
       self.pending_count[job["id"]] = len(paths)
       dls = job["dls_json"]
       format_name = dls["acquisition"]["format"]["name"].lower() if "format" in dls["acquisition"] and "name" in dls["acquisition"]["format"] else None

       for file_path in paths:
          if format_name == "csv" or file_path.endswith('.csv'):
              self._frcsv_proxy.read_format_start(job, file_path)
          elif format_name == "xml" or file_path.endswith('.rdf') or file_path.endswith('.xml'):
              self._frxml_proxy.read_format_start(job, file_path)
          elif format_name == "xls" or file_path.endswith(".xls") or file_path.endswith('.xlsx'):
              self._frxls_proxy.read_format_start(job, file_path)
          elif format_name == "json" or file_path.endswith('.json') or file_path.endswith('.json.gz'):
              self._frjson_proxy.read_format_start(job, file_path)
          else:
              raise RuntimeError(f'Unsupported format in {file_path}')

    def read_format_end(self, job, path, df):
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = self.pending_count[job["id"]] - 1 
        
        self.job_df[job["id"]][path] = df 
        if self.pending_count[job["id"]] == 0:
            self.update_job_step(job, 'read_format_ended', 0.25)

    def send_to_unarchive(self, paths, job):
        filter_paths = job['dls_json']["acquisition"]["decompress"]["path"]
        self.pending_count[job["id"]] = len(paths) * len(filter_paths) 
        for archive_path in paths:
          for filter_path in filter_paths:
            self._unarchive_proxy.unarchive(archive_path, filter_path, job)

    
    def unarchive_end(self, archive_path, filter_path, bytes, job):
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = self.pending_count[job["id"]] - 1
        dest_file = hashlib.sha1(f"{archive_path}/{filter_path}".encode()).hexdigest()+'_'+os.path.basename(filter_path)
        dir_in_staging = self.staging_path(job['id'], 'unarch')
        os.makedirs(dir_in_staging, exist_ok = True)
        path_in_staging = self.staging_path(job['id'], 'unarch', dest_file)
        self.decompressed_files[job['id']].append(path_in_staging)
        self._storage_proxy.write_file(path_in_staging, bytes, 'wb+').get()
        if self.pending_count[job["id"]] == 0:
            self.update_job_step(job, 'unarchive_ended', 0.15)

    def send_to_read_df(self, dfs, job):
        self.pending_count[job["id"]] = len(dfs)
        for path, df in dfs.items():
            self._dfreader_proxy.df2tuple(df, path, job, job["dls_json"])

    def read_df_end(self, tuples, n_tuples, issues, path, job): 
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = self.pending_count[job["id"]] - 1
        self.job_tuples[job["id"]][path] = tuples
        self.job_issues[job["id"]].extend(issues)
        
        progress =  (1.0/10) * len(self.job_tuples[job["id"]]) / (self.pending_count[job["id"]] + len(self.job_tuples[job["id"]]))
        
        if self.pending_count[job["id"]] == 0:
            self.write_issues(self.job_issues[job["id"]]) 
            errors_number = sum(issue['issue_severity']=='error' for issue in self.job_issues[job["id"]])
            warning_number = sum(issue['issue_severity']=='warning' for issue in self.job_issues[job["id"]])
            if errors_number == 0 and (n_tuples > 0 or warning_number == 0):
                self.update_job_step(job, 'read_df_ended', 0.3 + progress)
            else:
                self.fail_job(job)
        else:
            self.update_job_step(job, job["step"], 0.3 + progress)

    def send_to_standardize(self, tuples, job):
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = len(tuples)
        for path, ttuples in tuples.items():
            self._standardizer_proxy.standardize(ttuples, path, job, job["dls_json"])

    def standardize_end(self, tuples, n_tuples, issues, path, job): 
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = self.pending_count[job["id"]] - 1
        self.job_stdtuples[job["id"]][path] = tuples 
        # deleting existing ref-not-found issue types since they come from a previous execution
        self.job_issues[job["id"]] = [i for i in self.job_issues[job['id']] if i['issue_type'] != "ref-not-found"]
        self._storage_proxy.delete_db("issue", lambda i: int(i["job_id"]) == int(job['id']) and i['issue_type'] == "ref-not-found").get()
        
        self.job_issues[job["id"]].extend(issues)
        
        progress =  (1.0/10) * len(self.job_stdtuples[job["id"]]) / (self.pending_count[job["id"]] + len(self.job_stdtuples[job["id"]]))
        
        if self.pending_count[job["id"]] == 0:
            self.write_issues(self.job_issues[job["id"]]) 
            errors_number = sum(issue['issue_severity']=='error' for issue in self.job_issues[job["id"]])
            warning_number = sum(issue['issue_severity']=='warning' for issue in self.job_issues[job["id"]])
            if all(t is None for t in self.job_stdtuples[job['id']].values()):
                # If all tuples returnes are None means that some referential were missing and that this source needs to be delayed
                l.debug(f"Source {job['dls_json']['scope']['source']} has been reversed to read format end since some referential failed completely, probably some depending source is missing")
                self.update_job_step(job, 'read_df_ended',  0.4)
            elif errors_number == 0 and (n_tuples > 0 or warning_number == 0):
                self.update_job_step(job, 'standardize_ended',  0.4 + progress)
            else:
                self.fail_job(job)
        else:
            self.update_job_step(job, job["step"], 0.4 + progress)

    def send_to_annotate(self, tuples, job):
        self.pending_count[job["id"]] = len(tuples)
        for path, ttuples in tuples.items():
            self._nlp_proxy.annotate(ttuples, path, job)

    def annotate_end(self, tuples, path, job):
        if job["status"] == "failed":
          return
        self.pending_count[job["id"]] = self.pending_count[job["id"]] - 1
        self.job_stdtuples[job["id"]][path] = tuples
        if self.pending_count[job["id"]] == 0:
            self.update_job_step(job, 'annotate_ended', 0.55)

    def send_to_aggregate(self, tuples, job):
        self._aggregate_proxy.aggregate(tuples, job)

    def aggregate_end(self, tuples, job): 
        if job["status"] == "failed":
          return
        self.job_aggrtuples[job["id"]] = tuples
        self.update_job_step(job, 'aggregate_ended', 0.65)
    
    def send_to_precalculate(self, tuples, job):
        self._evaluator_proxy.plan_calculate(tuples, job)

    def precalculate_end(self, indicators_to_calculate, job): 
        if job["status"] == "failed":
          return
        self.job_precalinds[job["id"]] = indicators_to_calculate
        self.job_precalstep[job["id"]] = 0
        self.update_job_step(job, 'precalculate_ended', 0.75)

    def send_to_calculate(self, indicators_to_calculate, job): 
        self._evaluator_proxy.calculate(indicators_to_calculate, job)

    def calculate_end(self, ind_tuples, job): 
        if job["status"] == "failed":
          return
        self.job_indicators[job["id"]] = ind_tuples
        self.update_job_step(job, 'calculate_ended', 0.85)

    def send_to_publish(self, tuples, job):
        calc_step = self.job_precalstep[job['id']]
        self._variables_proxy.write_variable(tuples, calc_step, job)
    
    def publish_end(self, job): 
        if job["status"] == "failed":
          return
        self.update_job_step(job, 'publish_ended', 0.95)
    
    # this function returns a future that can be waited to ensure that file job is written to disk
    def update_job_step(self, job, step, progress):
        l.debug(f"Changing to step {step} for job {int(job['id'])} source {job['dls_json']['scope']['source']} {round(progress*100, 1)}%")
        # removing job from current step dict
        if job["step"] in self.job_steps and job["id"]:
          if job["id"] in self.job_steps[job["step"]]:
            self.job_steps[job["step"]].pop(job["id"])
          else:
            l.warning(f"Warning: could not find id {int(job['id'])} in job_steps {job['step']}") 
        # changing the job step 
        job["step"] = step
        job["progress"] = progress

        # addong the job to te new dict unless is on the last step
        if step == self.last_step:
          job["status"] = "success"
          job["end_on"] = datetime.datetime.now()
          job["progress"] = 1
          self._storage_proxy.delete_job_cache(job["id"]).get()
        else:
          self.job_steps[step][job["id"]] = job

        return self._storage_proxy.write_db(job, 'job')
    
    # this function returns a future that can be waited to ensure that file job is written to disk
    def fail_job(self, job, delete_job = False, issue = None):
        if type(job) == int :
          for stepjobs in self.job_steps.values():
            if job in stepjobs:
              job = stepjobs[job]
              break
        if type(job) == int:
          if delete_job:
            self.clean_job_ids([job])
          else:
            raise ValueError(f"Could not find job {job}")
        else:
          l.warning(f"Changing status of job {job['id']} to 'failed' for {job['dls_json']['scope']['source']}")
          # removing job from current step dict
          if job["step"] in self.job_steps and job["id"]:
            if job["id"] in self.job_steps[job["step"]]:
              self.job_steps[job["step"]].pop(job["id"])
            else:
              l.warning(f"Warning: could not find id {job['id']} in job_steps {job['step']}") 
          # changing the job status 
          job["status"] = "failed"

          if delete_job:
            self.remove_job(job)
          else:
            self._storage_proxy.write_db(job, 'job')
          # writing and issue if required
          if issue is not None:
            self._storage_proxy.write_db(record=issue, db_class='issue').get() 
          self._storage_proxy.delete_job_cache(job["id"]).get()

    def write_issues(self, issues):
        type_count = {issue['issue_type']:0 for issue in issues}
        for issue in issues:
            if type_count[issue['issue_type']] < 50:
              type_count[issue['issue_type']] =  type_count[issue['issue_type']] + 1
              self._storage_proxy.write_db(record=issue, db_class='issue').get() 

    def job_to_annotate(self, job):
      if "columns" in job['dls_json']:
        for col in job['dls_json']['columns']:
          if "variable" in col and col["variable"] == "article_text":
            return True
      return False

    def remove_job(self, job):    
        for job_dico in self.job_dicos:
            if job["id"] in job_dico:
              job_dico.pop(job["id"])
        self.clean_old_jobs(job["source"])

    def clean_old_jobs(self, source):
        # getting 10 oldest jobs
        jobs = self._storage_proxy.read_db("job", lambda j: j["source"] == source).get()
        if len(jobs) > self._jobs_to_keep:
          n_del = len(jobs) - self._jobs_to_keep
          to_del = jobs.sort_values("id", axis = 0, ascending = True).head(n_del)["id"].tolist()
          self.clean_job_ids(to_del) 
    
    def clean_job_ids(self, ids):
      to_del = {int(j) for j in ids}
      # deleting issues
      self._storage_proxy.delete_db("issue", lambda i: int(i["job_id"]) in to_del ).get()
      # deleting jobs
      self._storage_proxy.delete_db("job", lambda j: int(j["id"]) in to_del).get()
      # deleting staging
      for job_id in to_del:
        self._storage_proxy.delete_dir(self.staging_path(int(job_id)))




import os, re
import pandas as pd
import json
import shutil
from io import BytesIO
from . import worker
from . import util
import shelve

class Storage(worker.Worker):
   
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name, orchestrator_ref, settings)
     
    def loop_actions(self):
        pass

    def on_start(self):
        super().on_start()
        # create base directories
        if not os.path.exists(self.pandem_path("database")):
           os.makedirs(self.pandem_path("database"))
        if not os.path.exists(self.pandem_path("files")):
           os.makedirs(self.pandem_path("files"))
        if not os.path.exists(self.pandem_path("files", "staging")):
           os.makedirs(self.pandem_path("files", "staging"))
        # create dict of job cached objects
        # self.job_cache = {}
        # create empty dataframes in self.db_tables if pickle files doesn't exist
        self.db_tables = dict()
        test_bool = os.path.exists(os.path.join(os.getenv('PANDEM_HOME'), 'database/jobs.pickle'))
        if os.path.exists(os.path.join(os.getenv('PANDEM_HOME'), 'database/jobs.pickle')):
            self.db_tables['job'] = pd.read_pickle(os.path.join(os.getenv('PANDEM_HOME'), 'database/jobs.pickle'))
        else:
            self.db_tables['job'] = pd.DataFrame({'id': pd.Series(dtype='int'),
                                             'source': pd.Series(dtype='str'), 
                                             'source_files': pd.Series(dtype=object),
                                             'file_sizes': pd.Series(dtype=object),
                                             'progress': pd.Series(dtype=float),
                                             'start_on': pd.Series(dtype=object),
                                             'end_on': pd.Series(dtype=object),
                                             'step': pd.Series(dtype='str'), 
                                             'status': pd.Series(dtype='str'),
                                             'dls_json': pd.Series(dtype=object)
                                             })
        if os.path.exists(os.path.join(os.getenv('PANDEM_HOME'), 'database/issues.pickle')):
            self.db_tables['issue'] = pd.read_pickle(os.path.join(os.getenv('PANDEM_HOME'), 'database/issues.pickle'))
        else:
            self.db_tables['issue'] = pd.DataFrame({'id': pd.Series(dtype='int'),
                                             'step': pd.Series(dtype='str'), 
                                             'line': pd.Series(dtype='int'), 
                                             'source': pd.Series(dtype='str'),
                                             'file': pd.Series(dtype='str'),
                                             'message': pd.Series(dtype='str'),
                                             'raised_on': pd.Series(dtype=object),
                                             'job_id': pd.Series(dtype='int'), 
                                             'issue_type': pd.Series(dtype='str'),
                                             'issue_severity': pd.Series(dtype='str')
                                             })
        if os.path.exists(os.path.join(os.getenv('PANDEM_HOME'), 'database/sources.pickle')):
            self.db_tables['source'] = pd.read_pickle(os.path.join(os.getenv('PANDEM_HOME'), 'database/sources.pickle'))
        else:                                    
            self.db_tables['source'] = pd.DataFrame({'id': pd.Series(dtype='int'),
                                                     'name': pd.Series(dtype='str'), 
                                                     'last_hash': pd.Series(dtype='str'),
                                                     'last_exec':pd.Series(dtype=object),
                                                     'next_exec': pd.Series(dtype=object)
                                                    })   

             
    def write_file(self, path, bytes, mode): #absolute path here
        with open(path, mode) as f:
            f.write(bytes) 
    

    def read_file(self, path): #absolute path here
        if path.startswith('PANDEM_HOME'):
          file_path = path
        else :
          file_path = os.path.join(os.getenv('PANDEM_HOME'), 'files', path)
        
        if file_path.split('.')[-1] == "json":
            with open(file_path, 'r') as f:
                data_dict = json.load(f)
            return data_dict
        else :
            with open(file_path, 'rb') as f:
                bytes_data = BytesIO(f.read())
            return bytes_data


    def copy_files(self, src_paths, dest_paths): #absolute paths here
        if not os.path.isdir(os.path.dirname(dest_paths[0])):
            os.makedirs(os.path.dirname(dest_paths[0]))
        for (src_path, dest_path) in zip(src_paths, dest_paths):
            shutil.copyfile(src_path, dest_path)  
        
            
    def list_files(self, path, match=None, recursive=True, exclude=['.git'], include_dirs = False, include_files = True): #path under files
        files_paths = []
        if recursive:
            for (dirpath, dirnames, filenames) in os.walk(os.path.join(os.getenv('PANDEM_HOME'), 'files', path)):
                for dir in exclude:
                    if dir in dirnames:
                        dirnames.remove(dir)
                files_paths += [os.path.join(dirpath, file) for file in filenames]
        else:
            paths= os.listdir(os.path.join(os.getenv('PANDEM_HOME'), 'files', path))
            files_paths = [file_path for file_path in paths 
              if include_files and os.path.isfile(os.path.join(os.getenv('PANDEM_HOME'), 'files', path, file_path)) or
                 include_dirs and os.path.isdir(os.path.join(os.getenv('PANDEM_HOME'), 'files', path, file_path)) 

            ]
        if match is not None:
            matched_files = [{'path': file_path, 'name':os.path.basename(file_path)} for file_path in files_paths if re.mach(match, os.path.basename(file_path))] 
            return matched_files
        else:
            return [{'path': file_path, 'name':os.path.basename(file_path)} for file_path in files_paths]

    def list_files_abs(self, path, match=None, recursive=True, exclude=['.git']): #absolute path for target directory
        files_paths = []
        if recursive:
            for (dirpath, dirnames, filenames) in os.walk(path):
                for dir in exclude:
                    if dir in dirnames:
                        dirnames.remove(dir)
                files_paths += [os.path.join(dirpath, file) for file in filenames]
        else:
            paths= os.listdir(path)
            files_paths = [file_path for file_path in paths if os.path.isfile(os.path.join(path, file_path))]
        if match is not None:
            matched_files = [{'path': file_path, 'name':os.path.basename(file_path)} for file_path in files_paths if re.mach(match, os.path.basename(file_path))] 
            return matched_files
        else:
            return [{'path': file_path, 'name':os.path.basename(file_path)} for file_path in files_paths]
       
        
    def delete_file(self, path, match):
        if os.path.exists(path):
            files = os.listdir(os.path.join(os.getenv('PANDEM_HOME'), path))
            matched_files = [file for file in files if re.mach(match,file)]
            if len(matched_files) > 0:
                for file in matched_files:
                    os.remove(os.path.join(self.settings['home_dir'], path, file))
                return 'Ok'
            else:
                return 'no files founded' 
        else:
            raise FileNotFoundError("folder {0} does not exist!".format(path))  
    
    def exists(self, path):
        return os.path.exists(path)
    
    def delete_dir(self, path):
        if os.path.exists(path):
            shutil.rmtree(path)
        
    def write_db(self, record, db_class):
        df = self.db_tables[db_class]
        if not 'id' in  record:
            if df.shape[0] > 0:
                record['id'] = df.index.max()+1
            else:
                record['id'] = 1
            df.at[int(record['id'])] = record
        else:
            for key, value in record.items():
                df.at[int(record['id']), key] = value
        self.db_tables[db_class] = df
        dest = os.path.join(os.getenv('PANDEM_HOME'), 'database', db_class+'s'+'.pickle') 
        tmp = f"{dest}.tmp"
        df.to_pickle(tmp)
        shutil.move(tmp, dest)
        return record['id']


    def read_db(self, db_class, filter=None):
        df = self.db_tables[db_class]
        if df.shape[0] > 0:
            if filter != None:
                df = df.loc[df.apply(filter, axis = 1)]
            return df
        else:
            return None


    def delete_db(self, db_class, filter=None):
        df = self.db_tables[db_class]
        if filter != None and len(df) > 0:
            to_del = df.loc[df.apply(filter, axis = 1)]
            if len(to_del)>0:
              df = df.drop(index = to_del.index)
        self.db_tables[db_class] = df
        df.to_pickle(os.path.join(os.getenv('PANDEM_HOME'), 'database', db_class+'s'+'.pickle'))
        return df

    def get_job_cache(self, job_id):
      job_id = int(job_id)
      shelve_path = util.pandem_path("files", "staging", str(job_id), "cache")
      #if job_id in self.job_cache:
      #  self.job_cache[job_id] =  shelve.open(shelve_path,  writeback=False)
      #else:
      if not os.path.exists(util.pandem_path("files", "staging", str(job_id))):
          os.makedirs(util.pandem_path("files", "staging", str(job_id)))
      #  self.job_cache[job_id] =  
      return shelve.open(shelve_path,  writeback=False)
      #return self.job_cache[job_id]

    def to_job_cache(self, job_id, key, data):
      job_id = int(job_id)
      key = str(key)
      with self.get_job_cache(job_id) as cache:
        cache[key] = data
      return CacheValue(job_id, key, self._self_proxy)

    def from_job_cache(self, job_id, key):
      job_id = int(job_id)
      key = str(key)
      with self.get_job_cache(job_id) as cache:
        return cache[key]

    def delete_job_cache(self, job_id):
      job_id = int(job_id)
      shelve_path = util.pandem_path("files", "staging", str(job_id), "cache")
      if os.path.exists(shelve_path):
        os.remove(shelve_path)

   
 
class CacheValue:
  def __init__(self, job_id, key, storage_proxy):
    self.job_id = job_id
    self.key = key
    self._storage_proxy = storage_proxy

  def value(self):
    return self._storage_proxy.from_job_cache(self.job_id, self.key).get()





from . import acquisition
from . import util
import tweepy
import threading
import time
import json
import gzip
from datetime import datetime
import os
from math import inf
import re
import logging

l = logging.getLogger("pandem.twitter")

class AcquisitionTwitter(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "twitter")
    
    def on_start(self):
        super().on_start()
        # Getting twitter credentials
        self._api_key=util.get_or_set_secret("twitter-api-key") 
        self._api_key_secret=util.get_or_set_secret("twitter-api-key-secret") 
        self._access_token=util.get_or_set_secret("twitter-access-token") 
        self._access_token_secret=util.get_or_set_secret("twitter-access-token-secret") 
        self._filter_dir = util.pandem_path("files", "twitter", "v1.1", "tweets")
        self._filter_arc_dir = util.pandem_path("files", "twitter", "v1.1", "archived")
        if not os.path.exists(self._filter_dir):
          os.makedirs(name = self._filter_dir)
        if not os.path.exists(self._filter_arc_dir):
          os.makedirs(name = self._filter_arc_dir)

    def add_datasource(self, dls, force_acquire):
      if len(self.current_sources) > 0:
        raise ValueError("Twitter aqquisition support only a singlr DLS, others will be ignored")
      if "acquisition" in dls and "channel" in dls["acquisition"] and "topics" in dls["acquisition"]["channel"]:
        self._topics  = dls["acquisition"]["channel"]["topics"].keys()
        self._maingroup  = dls["acquisition"]["channel"]["main_group"]
        self._phrases = []
        self._included_regex = {}
        self._topic_groups = {}
        for topic in self._topics:
          self._included_regex[topic] = ""
          if "phrases" in dls["acquisition"]["channel"]["topics"][topic]:
            phrases = dls["acquisition"]["channel"]["topics"][topic]["phrases"]
            in_maingroup = dls["acquisition"]["channel"]["topics"][topic]["group"] == self._maingroup
            self._included_regex[topic] = "|".join(map(lambda v: re.escape(v.lower()), phrases))  
            for kw in phrases:
              if not kw in self._phrases and in_maingroup:
                if len(kw.encode("utf-8")) > 60:
                  raise ValueError(f"Twitter filter endpoint cannot contain phrases bigger than 60 bytes and {kw} has {len(kw.encode('utf-8'))}")
                self._phrases.append(kw)
          else:
            raise ValueError("Twitter DLS topics needs to have a phrases property with a list of phrases to use")
          if "group" in dls["acquisition"]["channel"]["topics"][topic]:
            self._topic_groups[topic] =  dls["acquisition"]["channel"]["topics"][topic]["group"]
      else: 
        raise ValueError("If a Twitter DLS is found it should contain a (possible empty) list of topics under ['acquisition']['channel']")
      
      if "acquisition" in dls and "channel" in dls["acquisition"] and "languages" in dls["acquisition"]["channel"]:
        self._languages = dls["acquisition"]["channel"]["languages"]
      else:
        self._languages = None

      if "acquisition" in dls and "channel" in dls["acquisition"] and "excluded_phrases" in dls["acquisition"]["channel"]:
        self._excluded_regex =  "|".join(map(lambda v: re.escape(v.lower()), dls["acquisition"]["channel"]["excluded_phrases"]))
      else:
        self._excluded_regex = None

      if "acquisition" in dls and "channel" in dls["acquisition"] and "include_retweets" in dls["acquisition"]["channel"]:
        self._include_retweets = bool(dls["acquisition"]["channel"]["include_retweets"])
      else:
        self._include_retweets = False
      
      # launching the tweet collection
      self.tweet_filter = self.TwitterFilter(
        track = self._phrases,
        included_regex = self._included_regex,
        excluded_regex = self._excluded_regex,
        include_retweets = self._include_retweets,
        topic_groups = self._topic_groups,
        languages = self._languages,
        filter_dir = self._filter_dir,
        consumer_key = self._api_key, 
        consumer_secret = self._api_key_secret, 
        access_token = self._access_token, 
        access_token_secret = self._access_token_secret
      )
      self.create_new_gz()
      threading.Thread(target=self.tweet_filter.run).start()
      super().add_datasource(dls, force_acquire)

    def create_new_gz(self):
       lfile = f"{datetime.now().strftime('%Y.%m.%d.%H.%M.%S')}.json.gz"
       path = util.pandem_path("files", "twitter", "v1.1", "tweets", lfile)
       l.info(f"Creating new file {lfile} for storing new tweets")
       open(path, 'a').close()
       return lfile
    
    def new_files(self, dls, last_hash):
        existing_files = list(filter(lambda f: f.endswith(".json.gz"), os.listdir(self._filter_dir)))
        # files to archive which are those with a name alphabetically lesser or equal than the current hash
        if last_hash is not None and last_hash !=  "":
          to_archive = list(filter(lambda v: v <=last_hash, existing_files))
        else: 
          to_archive = []
       
        # deleting files already processed
        for to_arc in to_archive:
          arc = os.path.join(self._filter_arc_dir, to_arc[0:10])
          os.makedirs(arc, exist_ok = True)
          os.rename(os.path.join(self._filter_dir, to_arc), os.path.join(arc, to_arc))
          l.debug(f"old file {to_arc} has been archived")

        # files to pipeline are thise with a name alphabetically bigger than the current hash
        files_to_pipeline = list(filter(lambda v: v > last_hash and os.path.getsize(os.path.join(self._filter_dir, v)) > 0, existing_files))
        files_to_pipeline.sort(reverse = False) 
        if len(files_to_pipeline) > 0:
          current_hash = files_to_pipeline[-1]
        else : 
          current_hash = ""
        # adding full path
        files_to_pipeline = list([os.path.join(self._filter_dir, f) for f in files_to_pipeline])
        # creating the new file and waiting until a first write is done to be sure than the previous files are not updated anymore
        new_name = self.create_new_gz()
        new_file = os.path.join(self._filter_dir, new_name)
        new_size = os.path.getsize(new_file)
        while new_size == 0:
          l.debug("Waiting 1 second to see if the new tweet file is being filled")
          time.sleep(1)
          new_size = os.path.getsize(new_file)
        l.debug("New tweet file contais data. Sending previous files to pipeline")
          
        return {"hash":current_hash, "files":files_to_pipeline}  

    class TwitterFilter(tweepy.Stream):
      def __init__(self, track, included_regex, excluded_regex, include_retweets, topic_groups, languages, filter_dir,
            consumer_key, consumer_secret, access_token, access_token_secret, *args, chunk_size=512, max_retries=inf, proxy=None, verify=True
          ): 
        self._track = track
        self._included_regex = included_regex
        self._excluded_regex = excluded_regex
        self._include_retweets = include_retweets
        self._topic_groups = topic_groups
        self._languages = languages
        self._filter_dir = filter_dir
        super().__init__(consumer_key, consumer_secret, access_token, access_token_secret, *args, chunk_size=chunk_size, daemon=False, max_retries=max_retries, proxy=proxy, verify=verify)

      def run(self):
        try:
          while(True) :
            l.debug("Launching twitter track")
            l.debug(f"tracking: {self._track}")
            l.debug(f"languages{self._languages}")
            l.debug(f"included regex {self._included_regex}")
            l.debug(f"excluded regex {self._excluded_regex}")
            l.debug(f"topic groups {self._topic_groups}")
            l.debug(f"include retweets {self._include_retweets}")
            self.filter(follow=None, track=self._track, locations=None, filter_level=None, languages=self._languages, stall_warnings=False)
            l.warning("Tweet filter failed, trying again in 1 minute")
            time.sleep(60)
        except Exception as e: 
          l.error(str(e))
      
      def on_status(self, status):
        if hasattr(status, "retweeted_status"):
          if not self._include_retweets:
            text = None
          elif hasattr(status.retweeted_status, "extended_tweet") and "full_text" in status.retweeted_status.extended_tweet:
            text = status.retweeted_status.extended_tweet["full_text"]
          else: 
            text = status.retweeted_status.text
        elif hasattr(status, "extended_tweet") and "full_text" in status.extended_tweet:
          text = status.extended_tweet["full_text"]
        else: 
          text = status.text
        if text is not None:
          file_name = self.get_gz_file()
          file_path = os.path.join(self._filter_dir, file_name)
          reporting_date = str(datetime.strptime(file_name, '%Y.%m.%d.%H.%M.%S.json.gz'))
          res = {
            "id":status.id_str,
            "created_at":str(status.created_at),
            "lang":status.lang,
            "text":text,
            "reporting_time": reporting_date,
            "article_count": 1
          }
          
          for t in self.matching_topics(text):
            if t in self._topic_groups:
              col_name = f"topic_{self._topic_groups[t]}"
            else:
              col_name = "topic"
            if col_name in res:
              res[col_name].append(t)
            else:
              res[col_name] = [t]

          with gzip.GzipFile(file_path, 'a') as fout:
            fout.write(f"{json.dumps(res)}\n".encode('utf-8')) 

      def get_gz_file(self):
        files = list(filter(lambda f: f.endswith(".json.gz"), os.listdir(self._filter_dir)))
        if len(files)==0:
          raise ValueError("Could not find any file to save the tweets")
        else:
          files.sort(reverse=True)
          return files[0]
     
      def matching_topics(self, text):
        return list(filter(
          lambda topic: re.search(self._included_regex[topic], text.lower()) is not None and re.search(self._excluded_regex, text.lower()) is None,
          self._included_regex.keys()
        ))




import sys, os
import argparse
import logging
import shutil
from . import util 
from .orchestrator import Orchestration
from . import admin
import pandas as pd

l = logging.getLogger("pandem")

def main(a):
  #conf = config()
  # Base argument parser
  parser = argparse.ArgumentParser()
  subs = parser.add_subparsers()
  
  # Launch pandem source start command
  start_parser = subs.add_parser("start", help = "Starts PANDEM source monitor")
  
  start_parser.add_argument(
    "-d", 
    "--debug", 
    action="store_true", 
    help="Whether to output debugging messages to the console", 
  )
  
  start_parser.add_argument(
    "--no-acquire", 
    action="store_true", 
    help="Prevent data acquisition from sources", 
  )
  
  start_parser.add_argument(
    "--no-nlp", 
    action="store_true", 
    help="Prevent NLP models and sources using them (Twitter, MediSys) pf being launched (Useful for instances without docker" 
  )

  start_parser.add_argument(
    "--retry-failed", 
    action="store_true", 
    help="Whether to retry failed jobs" 
  )
  
  start_parser.add_argument(
    "--not-retry-active", 
    action="store_true", 
    help="Whether to do not retry active jobs" 
  )
  start_parser.add_argument(
    "--no-app", 
    action="store_true", 
    help="If present the pandem2-source app will not be lauched" 
  )
  start_parser.add_argument(
    "--force-acquire", 
    action="store_true", 
    help="Reset the concerned data sources to force a fresh acquire" 
  )

  start_parser.add_argument(
    "-r",
    "--restart-job", 
    type=int, 
    required = False,
    default = 0,
    help="Job id to re-run if its data is still stored" 
  )
  
  start_parser.add_argument(
    "-l",
    "--limit-collection", 
    help="Comma separated list of sources for limiting the collection" 
  )
  
  start_parser.set_defaults(func = do_start)
  
  # setup 
  reset_parser = subs.add_parser("reset", help = "reset configuration as system defaults")
  
  reset_parser.add_argument(
    "-v", 
    "--variables", 
    action="store_true", 
    help="Whether to restore variables defiitions to last system defaults", 
  )
  reset_parser.add_argument(
    "--restore-factory-defaults", 
    action="store_true", 
    help="Delete all files and database elements and restore system defaults", 
  )
  reset_parser.add_argument(
    "--scripts",
    action="store_true",
    help="Restore the default script folder in pandem-home"
  )
  reset_parser.add_argument(
    "--covid19-datahub", 
    action="store_true", 
    help="Reset covid19-datahub datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--ecdc-atlas", 
    action="store_true", 
    help="Reset ecdc-atlas datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--influenzanet", 
    action="store_true", 
    help="Reset influenza net datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--ecdc-covid19", 
    action="store_true", 
    help="Reset ecdc-covid19 datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--ecdc-covid19-simulated", 
    action="store_true", 
    help="Reset ecdc-covid19 simulated datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--serotracker", 
    action="store_true", 
    help="Reset serotracker datasource to system defaults", 
  )
  reset_parser.add_argument(
    "--pandem-partners-template", 
    action="store_true", 
    help="Reset pandem partner templates to system defaults", 
  )
  reset_parser.add_argument(
    "--twitter", 
    action="store_true", 
    help="Reset pandem twitter system defaults", 
  )
  reset_parser.add_argument(
    "--medisys", 
    action="store_true", 
    help="Reset pandem medisys system defaults", 
  )
  reset_parser.add_argument(
    "--flights",
    action="store_true",
    help="Reset pandem flight related information to system defaults", 
  )

  reset_parser.set_defaults(func = do_reset)

  # Launch pandem source listt command
  start_parser = subs.add_parser("list", help = "List elements on this Pandem-Source instance")
  start_parser.add_argument(
    "--sources",
    action="store_true",
    help="List existing sources in this instance", 
  )
  
  start_parser.add_argument(
    "--missing-sources",
    action="store_true",
    help="List sources embedded on pandemSource package not present on this instance", 
  )
  start_parser.add_argument(
    "--package-sources",
    action="store_true",
    help="List sources embedded on pandemSource package", 
  )
  start_parser.add_argument(
    "--missing-package-sources",
    action="store_true",
    help="List existing sources in this instance not present on pandemSource package", 
  )

  start_parser.set_defaults(func = do_list)

  
  util.check_pandem_home()
  #calling handlers
  func = None
  try:
    args = parser.parse_args()
    func = args.func
  except AttributeError:
    parser.print_help()
  if func != None:
    args.func(args, parser)

# handlers
def do_start_dev(debug = True, no_acquire = False, retry_failed = False, restart_job = 0, not_retry_active = False, no_app = True, force_acquire = False, no_nlp = False):
  from types import SimpleNamespace
  return do_start(SimpleNamespace(**{"debug":True, "no_acquire":no_acquire, "retry_failed":retry_failed, "limit_collection":None, "restart_job":restart_job, "no_app":no_app, "not_retry_active":not_retry_active, "force_acquire":force_acquire, "no_nlp": no_nlp}))

# handlers
def do_start(args, *other):
  if args.debug:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)
 
  pkg_dir, this_filename = os.path.split(__file__)
  defaults = os.path.join(pkg_dir, "data/defaults.yml") 
  config = util.pandem_path("settings.yml")
  if not os.path.exists(config):
    shutil.copy(defaults, config)

  # Adding python scripts on pandem_home to the system path
  sys.path.insert(1, util.pandem_path("files", "scripts", "py"))
  settings = util.settings()
  
  install_issues = admin.install_issues(check_nlp = not args.no_nlp)
  if len(install_issues) > 0:
    eol = '\n'
    l.warning(f"""The following errors where found on current PANDEM-2 installation please fix before proceed
    {eol.join(install_issues)}
    """)
    return None
  
  if args.restart_job > 0:
    orchestrator_ref = Orchestration.start(settings, start_acquisition = False, retry_failed = False, restart_job = args.restart_job, retry_active = True, force_acquire = args.force_acquire, no_nlp = args.no_nlp)
    orch = orchestrator_ref.proxy()
  elif args.limit_collection is not None:
    orchestrator_ref = Orchestration.start(settings, start_acquisition = False, retry_failed = args.retry_failed, restart_job = args.restart_job, retry_active = not args.not_retry_active, force_acquire = args.force_acquire, no_nlp = args.no_nlp)
    orch = orchestrator_ref.proxy()
    storage_proxy = orch.get_actor("storage").get().proxy()
    dls_files = storage_proxy.list_files('source-definitions').get()
    dls_dicts = [storage_proxy.read_file(file_name['path']).get() for file_name in dls_files if file_name['path'].endswith(".json")]
    
    for source_name in args.limit_collection.split(","):
      dls = list(filter(lambda dls: dls['scope']['source'] == source_name, dls_dicts))[0]
      acquisition_proxy = orch.get_actor(f"acquisition_{dls['acquisition']['channel']['name']}").get().proxy()
      acquisition_proxy.add_datasource(dls, args.force_acquire)
  else:
    orchestrator_ref = Orchestration.start(settings, start_acquisition = not args.no_acquire, retry_failed = args.retry_failed, retry_active = not args.not_retry_active, force_acquire = args.force_acquire, no_nlp = args.no_nlp)
    orch = orchestrator_ref.proxy()

  # launching pandem2source app
  if not args.no_app:
    admin.run_pandem2app()    

  return orch
  
def do_reset(args, *other):
  if args.restore_factory_defaults:
    admin.delete_all()
    admin.reset_default_folders("input-local", "input-local-defaults", "dfcustom", "scripts", "variables", "indicators", "img")
  if args.scripts:
    admin.reset_default_folders("scripts")
  if args.variables or args.restore_factory_defaults:
    admin.reset_variables()
  if args.covid19_datahub or args.ecdc_covid19 or args.restore_factory_defaults:
    admin.reset_source("nuts-eurostat")
    admin.reset_source("ICD-10-diseases-list")
  if args.pandem_partners_template or args.restore_factory_defaults:
    admin.reset_source("covid19-template-cases")
    admin.reset_source("covid19-template-cases-RIVM")
    admin.reset_source("covid19-template-cases-THL")
    admin.reset_source("covid19-template-beds-staff")
    admin.reset_source("covid19-template-contact-tracing")
    admin.reset_source("covid19-template-daily-voc-voi")
    admin.reset_source("covid19-template-daily-voc-voi-RIVM")
    admin.reset_source("covid19-template-hospitalised-deaths")
    admin.reset_source("covid19-template-hospitalised-deaths-THL")
    admin.reset_source("covid19-template-hospitalised-deaths-RIVM")
    admin.reset_source("covid19-template-ltcf-RIVM")
    admin.reset_source("covid19-template-outbreaks")
    admin.reset_source("covid19-template-participatory-sentinel-surveilla")
    admin.reset_source("covid19-template-sex-age-comorbidity")
    admin.reset_source("covid19-template-testing")
    admin.reset_source("covid19-template-testing-THL")
    admin.reset_source("covid19-template-testing-RIVM")
    admin.reset_source("covid19-template-vaccination")
    admin.reset_source("covid19-template-vaccination-THL")
    admin.reset_source("covid19-template-voc-voi")
    admin.reset_source("covid19-template-local-regions")
    admin.reset_source("covid19-template-local-regions-RIVM")
    admin.reset_source("covid19-template-local-regions-THL")
  if args.covid19_datahub or args.restore_factory_defaults:
    admin.reset_source("covid19-datahub")
  if args.ecdc_covid19_simulated or args.restore_factory_defaults:
    admin.reset_source("ecdc-covid19-age-group-variants")
  if args.ecdc_covid19 or args.restore_factory_defaults:
    admin.reset_source("ecdc-covid19-variants")
    admin.reset_source("ecdc-covid19-age-group")
    admin.reset_source("ecdc-covid19-measures")
    admin.reset_source("ecdc-covid19-daily")
  if args.serotracker or args.restore_factory_defaults:
    admin.reset_source("geonames-countries")    
    admin.reset_source("serotracker")    
  if args.ecdc_atlas or args.restore_factory_defaults:
    admin.reset_source("ecdc-atlas-influenza")    
  if args.influenzanet or args.restore_factory_defaults:
    admin.reset_source("influenza-net")    
  if args.twitter or args.restore_factory_defaults:
    admin.reset_source("twitter")    
  if args.medisys or args.restore_factory_defaults:
    admin.reset_source("medisys")
  if args.flights or args.restore_factory_defaults:
    admin.reset_source("ourairports")
    admin.reset_source("opensky-network-coviddataset")

def do_list(args, *other):
  if args.sources or args.missing_sources or args.package_sources or args.missing_package_sources:
    sources = admin.list_sources(local = args.sources, default = args.package_sources, missing_local = args.missing_sources, missing_default = args.missing_package_sources)
    for line in pd.DataFrame(sources,columns=["Source", "Tag"]).to_string().split("\n"):
        print(line)



if __name__ == "__main__":
  main(sys.argv[1] if len(sys.argv)>1 else None)





import os
import hashlib
from . import acquisition


class AcquisitionLocalFS(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "input-local")
        
    def new_files(self, dls, last_hash):
        file_path = self.source_path(dls, dls['acquisition']['channel']['xls_file'])
        if os.path.exists(file_path):
            current_hash = hashlib.md5(open(file_path,'rb').read()).hexdigest()
            files_to_pipeline = []
            if not os.path.exists(file_path) or last_hash == "":
                files_to_pipeline.extend([file_path])
            # the file already exists and we know the last etag 
            elif current_hash != last_hash:
                files_to_pipeline.extend([file_path])   
            return {"hash":current_hash, "files":files_to_pipeline}  
        else:
            return {"hash":'', "files":[]} 


from . import formatreader
import pandas as pd

class FormatReaderCSV(formatreader.FormatReader):

    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
    
    def read_df(self, file_path, dls):
        file_bytes = self._storage_proxy.read_file(file_path).get()
        params = {"filepath_or_buffer":file_bytes}
        if "decimal_sign" in dls["acquisition"]["format"]: 
          params["decimal"] =  dls["acquisition"]["format"]["decimal_sign"]
        if "skiprows" in dls["acquisition"]["format"]:
          params["skiprows"] =  dls["acquisition"]["format"]["skiprows"]
        if "sep" in dls["acquisition"]["format"]:
          params["sep"] =  dls["acquisition"]["format"]["sep"]
        if "thousands_separator" in dls["acquisition"]["format"] and len(dls["acquisition"]["format"]["thousands_separator"]) == 1:
          params["thousands"] =  dls["acquisition"]["format"]["thousands_separator"]
        if "encoding" in dls["acquisition"]["format"]:
          params["encoding"] =  dls["acquisition"]["format"]["encoding"]
        
        if file_bytes != '':
            df = pd.read_csv(**params) 
        else:
            raise ValueError('can not read from empty bytes')
        return df
    
    


          
   



import os

from numpy import int64
from . import worker
from abc import ABC, abstractmethod, ABCMeta
from datetime import datetime, timedelta
import time
import re
from collections import defaultdict
import itertools as it
import tempfile
import json
import subprocess
from copy import deepcopy
from itertools import chain
import logging
import traceback

l = logging.getLogger("pandem.evaluator")






class Evaluator(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)    
         
    def on_start(self):
        super().on_start()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy() 
        self._variables_proxy = self._orchestrator_proxy.get_actor('variables').get().proxy() 
        self._dict_of_variables = self._variables_proxy.get_variables().get()
        self._indicators, self._modifiers, self._parameters, self._scripts, self._params_set = self.get_indicators(self._dict_of_variables)

    def get_indicators(self, var_dic):
        # Read the formulas of all indicators 
        formulas = {var:var_dic['formula'] for var, var_dic in var_dic.items() if var_dic['formula']} 
        # Parse parameters from the functions/formula text and store them in parameters dict: for each indicator, returning a list for each parameters. e.g. parameters[“incidence”]=[“reporting_period”, “number_of_cases”, “population”]
        parameters = {ind:re.findall (r'([^(, )]+)(?!.*\()', formula) for ind, formula in formulas.items()}
        modifiers = {ind:({m["variable"]:m["value"] for m in var_dic[ind]["modifiers"]}) if "modifiers" in var_dic[ind] else {} for ind in var_dic}
        # Build indicators dict that returns for each variable used on an indicator a list of indicators where this variable is used. eg. indicators[“population”]=[“Incidence”, “prevalence, …”]
        scripts = {ind:re.search(r'^.*?(?=\()', formula).group() for ind, formula in formulas.items()}
        indicators = dict()
        params_set = {params[i] for ind, params in parameters.items() for i in range(len(params))}
        for param in params_set:
            ind_list = [ind for ind, params in parameters.items() if param in params]
            indicators[param] = ind_list

        return indicators, modifiers, parameters, scripts, params_set

    #def get_obs_name(self, obs, attrs):
    #    if self._dict_of_variables[obs]["aliases"] is None or len(self._dict_of_variables[obs]["aliases"]) == 0 or self._dict_of_variable[obs]["variable"] != obs:
    #      return obs
    #    else:
    #      for alias in self._dict_of_variables[obs]["aliases"]:
    #        if all(modifier["variable"] in attrs and attrs[modifier["variable"]] == modifier["value"] for modifier in alias["modifiers"]):
    #          return alias["alias"]
    #    return obs
    #          

    def modifiers_in_key(self, var_name, tuple_key):
      var_dic = self._dict_of_variables
      if "modifiers" in var_dic[var_name]:
        for mod in modifiers:
          mod_found = False
          for var, value in tuple_key.items():
            if var == mod["variable"]:
              mod_found = True
              if value != mod["value"]:
                return False
              break
          if not mod_found:
            return False
      return True

    def plan_calculate(self, list_of_tuples, job):
        # getting tuples from cache
        list_of_tuples = list_of_tuples.value()['tuples']
        
        var_dic = self._dict_of_variables
        modifiers = self._modifiers
        params_set = self._params_set
        parameters = self._parameters
        obs_keys = {}
        obs_aliases = {}
        next_keys = {}
        for t in list_of_tuples:
          if "obs" in t and "attrs" in t: 
            base_name = next(iter(t["obs"].keys()))
            var_names = [base_name]
            # evaluating if current tuple is a derived variable (to allow proper detection of derivated user provided variables)
            if 'aliases' in  var_dic[base_name]:
              for alias in var_dic[base_name]['aliases']:
                if all(m['variable'] in t['attrs'] and t['attrs'][m['variable']] == m['value'] for m in alias['modifiers']):
                  var_names.append(alias['alias'])
            
            for var_name in var_names:
              if var_name in params_set or var_name in parameters.keys():
                if var_name not in obs_keys:
                  obs_keys[var_name] = {
                    "comb":set(),
                    "dates":set()
                  }
                date_attrs = set(vn for vn in t["attrs"].keys() if var_dic[vn]['type'] == 'date' and t["attrs"][vn] is not None)
                independent_attrs = [t for t in t["attrs"].keys() if t not in modifiers[var_name] or  modifiers[var_name][t] is not None]
                sorted_attrs = list(sorted(vn for vn in independent_attrs if var_dic[vn]['type'] not in ['not_characteristic', 'date', 'private'] and t["attrs"][vn] is not None))
                if len(sorted_attrs) > 0 and len(date_attrs)>0:
                  obs_keys[var_name]["comb"].add(tuple((vn, t["attrs"][vn]) for vn in sorted_attrs))
                  obs_keys[var_name]["dates"].add(tuple((vn, t["attrs"][vn]) for vn in date_attrs))


        var_obs = {}

        indicators_to_cal = {}
        step = 0
        stop = False
        while not stop:
            stop = True
            for ind, params in parameters.items():
                # not trying a varibale being already present
                if not ind in obs_keys:
                    # testing the tuples than satisfy the provided parameters
                    mod = modifiers[ind]
                    date_par = next(p for p in params if var_dic[p]['type'] == 'date')
                    base_date = var_dic[date_par]['variable']
                    no_date_pars = [p for p in params if p != date_par]
                    attr_pars =  [p for p in no_date_pars if var_dic[p]['type'] not in {'observation', 'indicator', 'resource'}]
                    obs_pars =  list([p for p in params if var_dic[p]['type'] in {'observation', 'indicator', 'resource'}])
                    base_pars =  list([var_dic[p]["variable"] for p in obs_pars])
                    # in order to test this indicator we need to find at least the first observation on the current values
                    if len(obs_pars) > 0 and  base_pars[0] in obs_keys:
                      main_obs = obs_pars[0]
                      main_base = base_pars[0]
                      # We need to identify all tuples present on all obs_pars that respect the attr pars  
                      comb = list(obs_keys[main_base]["comb"])
                      dates = obs_keys[main_base]["dates"]
                      date_filter = {base_date: {v for date_comb in dates for k, v in date_comb if k == base_date}}
                      #date_filter.update(mofifiers[date_par])
                      # we can proceed the date_par has been found
                      if len(date_filter[base_date]) > 0:
                        # Iterating over all obs_par in formula
                        for i in range(0, len(obs_pars)):
                            obs_to_test = obs_pars[i]
                            base_to_test = base_pars[i]
                            # if the current obs base variable is not on obs key we have to look for it on published data
                            if len(comb) > 0 and base_to_test not in obs_keys:
                               # if the current obs has null mofifiers we have to delete those from lookup key
                               pub_comb = self._variables_proxy.lookup([base_to_test], comb, job['dls_json']['scope']['source'], date_filter, include_source = False, include_tag = True).get()
                               obs_keys[base_to_test] = {
                                 "comb":set(pub_comb.keys()),
                                 "dates":dates
                               }
                            # iterating over current possible combinations
                            j = 0
                            #if ind == "incidence" and obs_to_test == "population":
                            #    breakpoint()
                            while j < len(comb):
                                # the current combination should exists on obs_keys for the current parameter base observation
                                # tuple must contain the attrs_pars unless the current observation overrides it with a modifier
                                attr_pars_ok = True
                                # testing independent keys which is attributes not modified by the variable
                                key = comb[j]
                                for attr_par in attr_pars:
                                  ignore_attr = attr_par in modifiers[obs_to_test]
                                  if not ignore_attr and not any(k == attr_par for k, v in key):
                                    attr_pars_ok = False
                                    break
                                if attr_pars_ok:
                                  # checking that tuple contains the modifiers of the current observation unless is null
                                  attrs_obs_ok = True
                                  for obs_mod_var, obs_mod_value in modifiers[obs_to_test].items():
                                    if obs_mod_value is not None:
                                      if not any(k == obs_mod_var and v == obs_mod_value for k, v in key):
                                        attrs_obs_ok = False
                                        break
                                  if attrs_obs_ok:
                                    # checking if the current combination exists for the expected base variable
                                    # exluding modifiers
                                    indep_key = tuple([(k, v) for k, v in key if k not in modifiers[obs_to_test] or modifiers[obs_to_test][k] is not None])
                                    if base_to_test in obs_keys and indep_key in obs_keys[base_to_test]["comb"]:
                                      j = j + 1
                                    else:
                                      comb.pop(j)
                                  else:
                                    comb.pop(j)
                                else:
                                  comb.pop(j)
                        # The remaining combination have passed all validations to calculate the candidate indicator
                        if len(comb) > 0:
                          next_keys[ind] = {
                            "comb":set(comb),
                            "dates":dates
                          }
                          # adding the found indicator to the list to calculate
                          indicators_to_cal[ind] = {
                            "step":step + 1,
                            "comb":set(comb),
                            "dates":dates,
                            "date_par":date_par
                          }
                          # we have found something new so we will try to performa a new step
                          stop = False
            step = step + 1
            obs_keys.update(next_keys)
            next_keys.clear()
        self._pipeline_proxy.precalculate_end(indicators_to_cal, job=job)

       
    def prepare_scripts(self, ind, job):
        scripts = self._scripts
        params = self._parameters
        vars_dic =  self._dict_of_variables
        
        # read the functin code from indcators directory
        function_path = self.pandem_path(f'files', 'indicators', scripts[ind], 'function.R')
        if not os.path.exists(function_path):
            raise FileNotFoundError(f"Cannot found R scrip for calculate indicator {ind} it should be located at {function_path}")
        with open(function_path) as f:
            function_code = f.readlines()
        staging_dir = self.staging_path(job['id'], f'ind/{ind}')
        if not os.path.exists(staging_dir):
            os.makedirs(staging_dir)
        exec_file_path = os.path.join(staging_dir, 'exec.R')
        result_path = os.path.join(staging_dir, 'result.json')
        if os.path.exists(result_path):
          os.remove(result_path)
        # write the R script within the exec.R file
        execf = open(exec_file_path, 'w+')
        for param in params[ind]:
            param_file_path = os.path.join(staging_dir, param+'.json')
            execf.write(f'`{param}_mat` <- jsonlite::fromJSON("{param_file_path}")'+'\n')
            if os.path.exists(param_file_path):
              os.remove(param_file_path)
        param0 = params[ind][0]
        execf.write('res <- sapply(1:ncol('+param0+'_mat), function(col) {'+'\n')
        for param in params[ind]:
          execf.write(f'`{param}` = {param}_mat[,col]'+'\n')
        for code_line in function_code:
          execf.write(code_line)
        execf.write('})\n')
        execf.write(f'jsonlite::write_json(res, "{result_path}", matrix = "columnmajor")'+'\n')
        execf.close()

    def calculate(self, indicators_to_cal, job):
        try:
            vars_dic =  self._dict_of_variables
            params = self._parameters
            source =  job['dls_json']['scope']['source']
            result = {"tuples": []}
            # mixing indicators to calculate on all paths
            if indicators_to_cal and len(indicators_to_cal) > 0:
                # looping trough all indicators
                for ind, ind_map in indicators_to_cal.items():
                    combis = ind_map["comb"]
                    l.debug(f"calculating {ind} in {source} for {len(combis)} combinations")
                    # creating scripts for indicator
                    self.prepare_scripts(ind, job)
                    # getting all matching combinations for the indicator calculation from the database
                    obs = {p:vars_dic[p]["variable"] for p in params[ind] if vars_dic[p]["type"] in ["observation", "indicator", "resource"]}
                    var_date, base_date = next(iter((p, vars_dic[p]['variable']) for p in params[ind] if vars_dic[p]["type"] == "date"))
                    main_par, main_base = next(iter((p, vars_dic[p]['variable']) for p in params[ind] if vars_dic[p]["type"] in ["observation", "indicator", "resource"] ))
                    attrs = {p:vars_dic[p]["variable"] for p in params[ind] if p not in set(obs.keys())}
                    # iterating though each combination and launching the scripts to calculate the results
                    if len(combis) > 0: 
                      data = self._variables_proxy.lookup(list(obs.values()), combis, source, {base_date:None} , include_source = True, include_tag = True).get()
                      # getting sorted dates
                      dates = sorted({v["attrs"][base_date] for row in data.values() for v in row[main_base] })
                      # writing parameters matrices 
                      staging_dir = self.staging_path(job['id'], f'ind/{ind}')
                      can_run = True
                      # on file per parameter 
                      for p in params[ind]:
                        param_file_path = os.path.join(staging_dir, p+'.json')
                        with open(param_file_path, 'w') as jsonf:
                          jsonf.write("[")
                          sep = ""
                          # one row per date
                          all_none = True
                          for date in dates:
                            # one element per combination
                            param_values = [self._get_param_value(p, comb, date, data, obs, attrs, main_par, base_date) for comb in combis]
                            jsonf.write(sep)
                            sep = ","
                            for v in param_values:
                              if v is not None:
                                all_none = False
                            # writing the values on the parameter file 
                            jsonf.write(json.dumps(param_values))
                          jsonf.write("]")
                          if all_none:
                            can_run = False

                      # executing the script
                      exec_file_path = os.path.join(staging_dir, 'exec.R')
                      result_path = os.path.join(staging_dir, 'result.json')
                      if can_run:
                        subprocess.run (f'Rscript --vanilla {exec_file_path}', shell=True, cwd=staging_dir)
                        if os.path.exists(result_path):
                            modifiers = {t["variable"]:t["value"] for t in vars_dic[ind]["modifiers"] } if "modifiers" in vars_dic[ind] else {}
                            with open(self.pandem_path(result_path)) as f:
                                r = json.load(f)
                            assert(len(r) == len(combis))
                            for combi_res, comb in zip(r, combis):
                              for date, value in zip(dates, combi_res):
                                ind_date_tuple = {'obs': {ind:value if value != "NA" else None},
                                                'attrs':{**{base_date:date, "source":source},
                                                         **{k:v for k,v in comb},
                                                         ** modifiers
                                                         }
                                                }
                                result['tuples'].append(ind_date_tuple)
                        else:
                            l.warning(f'result file {result_path} not found')
            
            result['scope'] = {}            
            result['scope']['update_scope'] = [{'variable':'source', 'value':[source]}]            

            ret = self._storage_proxy.to_job_cache(job["id"], f"calc", result).get()
            self._pipeline_proxy.calculate_end(ret, job = job).get()
        except Exception as e: 
            l.warning(f"Error during calculation, job {job['id']} will fail")
            for line in traceback.format_exc().split("\n"):
              l.debug(line)
            issue = { "step":job['step'], 
                   "line":0, 
                   "source":job['source'], 
                   "file":'', 
                   "message":f"{str(e)}\n{traceback.format_exc()}", 
                   "raised_on":datetime.now(), 
                   "job_id":job['id'], 
                   "issue_type":"error",
                   'issue_severity':"error"
            }
            self._pipeline_proxy.fail_job(job, issue = issue).get()

    def _get_param_value(self, param, comb, date, data, obs, attrs, main_par, base_date):
        if comb in data:
          row = data[comb]
          # if the parameter is an observation we will look into the associated row attribute getting the date from attrs 
          if param in obs:
            for v in row[obs[param]]:
              if v["attrs"][base_date] == date:
                return v["value"]
          # if not an observation then the result is expected to be on attrs or in combination
          else:
            for v in row[obs[main_par]]:
              if v["attrs"][base_date] == date:
                if attrs[param] in v["attrs"]:
                  return v["attrs"][attrs[param]]
                else:
                  comb_map = {k:v for k, v in comb if k == attrs[param]}
                  return comb_map[attrs[param]]
        return None




import os
import subprocess
from . import acquisition_git


class AcquisitionGITLocal(acquisition_git.AcquisitionGIT):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
    
    def on_start(self):
        super().on_start()
        self._script_executor_proxy = self._orchestrator_proxy.get_actor('script_executor').get().proxy()

    def new_files(self, dls, last_hash):
        script_name = dls["acquisition"]["channel"]["changed_by"]["script_name"]
        script_type = dls["acquisition"]["channel"]["changed_by"]["script_type"]
        working_dir = self.source_path(dls)
        # executing the updating script
        self._script_executor_proxy.execute(script_name = script_name, script_type = script_type, working_dir = working_dir).get()

        # creating repo if it does no exists
        # if not os.path.exists(working_dir):
        #     os.makedirs(working_dir)
        res = subprocess.run(['git', 'status'], cwd = working_dir)
        if res.returncode != 0 :
            res = subprocess.run(["git", "init"], cwd = working_dir)
            res = subprocess.run(["git", "config", "user.name", "pandem2monitor"], cwd = working_dir)
            res = subprocess.run(["git", "config", "user.email", "pandem2monitor@pandem2.eu"], cwd = working_dir)
            res = subprocess.run(["git", "add", "."], cwd = working_dir)
            res = subprocess.run(["git", "commit", "-m" "Automatically initialized repo"], cwd = working_dir, capture_output=True)
        else :
            res = subprocess.run(["git", "add", "."], cwd = working_dir)
            res = subprocess.run(["git", "commit", "-m" "Committed new files"], cwd = working_dir, capture_output=True)
        
        return super().new_files(dls, last_hash)



import time
from . import worker
from . import util
from abc import ABC, abstractmethod, ABCMeta
import logging
from datetime import datetime
import traceback

l = logging.getLogger("pandem.formatreader")

class FormatReader(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
         
    def on_start(self):
        super().on_start()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy()

    def loop_actions(self):
        pass

    @abstractmethod
    def read_df(self, file_path, dls):
        pass
    
    def read_format_start(self, job, file_path):
        try:
          #print(f'I m in read_format_start : {file_path} with job : {job}')
          df = self.read_df(file_path, job['dls_json'])
          dls = job['dls_json']

          # adding column line number if it does not exists
          if 'line_number' not in df.columns :
              df['line_number'] = range(1, len(df)+1) 

          # adding file_path column
          if 'file' not in df.columns:
            df['file'] = file_path

          # calling custom df transformer if defined
          custom_trans = util.get_custom(["sources", dls["scope"]["source"].replace("-", "_").replace(" ", "_")],"df_transform")
          if custom_trans is not None:
            df = custom_trans(df)

          # Applying filters if any
          if "filter" in dls:
            for f in dls["filter"]:
              if "type" in f and f["type"] == "in" and "column" in f and "values" in f and f["column"] in df:
                df = df[df[f["column"]].isin(f["values"])]

          # reseting index
          df.reset_index(drop=True, inplace=True)
          df = self._storage_proxy.to_job_cache(job["id"], f"df_{file_path}", df).get()
          self._pipeline_proxy.read_format_end(job, file_path, df).get()
        except Exception as e: 
          l.warning(f"Error during dataframe reading, job {job['id']} will fail")
          for line in traceback.format_exc().split("\n"):
            l.debug(line)

          issue = { "step":job['step'], 
                   "line":0, 
                   "source":job['source'], 
                   "file":file_path, 
                   "message":str(e), 
                   "raised_on":datetime.now(), 
                   "job_id":job['id'], 
                   "issue_type":"error",
                   'issue_severity':"error"
          }
          self._pipeline_proxy.fail_job(job, issue = issue).get()
        





from . import formatreader
import pandas as pd
import re
import json
import itertools
import gzip

class FormatReaderJSON(formatreader.FormatReader):

    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)

    def read_df(self, file_path, dls):
        end_lined = "new_line_ended" not in dls["acquisition"]["format"] or bool(dls["acquisition"]["format"]["new_line_ended"])
        encoding = dls["acquisition"]["format"]["encoding"] if "encoding" in dls["acquisition"]["format"] else "UTF-8"
        gzipped = file_path.endswith(".gz")

        with (open(file_path,  'r', encoding=encoding) if not gzipped else gzip.open(file_path, 'rb')) as f:
           if end_lined :
              items = list(map(lambda l: json.loads(l), f.readlines())) 
           else :
              items = json.loads(f.read())

        row_steps = dls["acquisition"]["format"]["row"] if "row" in dls["acquisition"]["format"] else "*"
        row_steps = list(filter(lambda v: len(v)>0, row_steps.split("/")))

        #iterating over rows
        col_paths = {col['name']:list(filter(lambda v: len(v)>0, col['path'].split("/"))) for col in dls['columns'] }
        cols = list(col_paths.keys())
        rows = []
        for row_index in self.expand(items, row_steps):
          #Evaluating values for each row
          col_values = {c:self.values(items, row_index, col_paths[c]) for c in cols}
          combinations = itertools.product(*(col_values[c] for c in cols))
          for t in combinations:
            rows.append(t)

        df = pd.DataFrame(rows, columns = cols)
        return df

    def apply_index(self, items, index):
      x = items
      for i in index:
        x = x[i]
      return x

    def expand(self, items, steps, index = []):
      current = self.apply_index(items, index)
      if len(index) == len(steps):
        yield index
      else:
        i_step = len(index)

        step = steps[i_step]
        if step == "*":
          if not isinstance(current, list):
            raise ValueError(f"Cannot apply json path {step} since the current object is not a list")
          else:
            for i in range(0, len(current)):
              for y in self.expand(items, steps, index + [i]):
                yield y
        elif step.startswith("@"):
          prop = step[1:len(step)]
          if prop not in current:
            raise ValueError(f"Cannot apply json path {step} since the current object does not contains a property with that name")
          else:
            for y in self.expand(items, row_steps, index + [prop]):
              yield y
        else:
          raise ValueError(f"Cannot understund path {step}")

    def values(self, items, index, paths):
      current = self.apply_index(items, index)
      if len(paths) == 0:
        yield current
      else:
        step = paths[0]
        rest = paths[1:len(paths)]
        if step == "*":
          if not isinstance(current, list):
            raise ValueError(f"Cannot apply json path {step} of {paths} on since the current object is not a list")
          else:
            for i in range(0, len(current)):
              for y in self.values(items, index + [i], rest): 
                yield y
        elif step.startswith("@"):
          prop = step[1:len(step)]
          if prop not in current:
            yield None
          else:
            for y in self.values(items, index + [prop], rest):
              yield y
        elif step == "parent()":
          if len(index)==0:
            raise ValueError(f"Cannot apply json path {step} of {paths} since the current object is root")
          else:
            for y in self.values(items, index[0:len(index)-1], rest):
              yield y
        else:
          raise ValueError(f"Cannot understund path {step}")



import os
import datetime
import json
import numpy
import io
import getpass
import yaml

def check_pandem_home():
  if os.environ.get("PANDEM_HOME") is None:
    raise RuntimeError("The variable PANDEM_HOME needs to be set to a local folder in order to run pandem source")

def pandem_path(*path):
  if os.environ.get("PANDEM_HOME") is None:
    raise RuntimeError("The variable PANDEM_HOME needs to be set to a local folder in order to run pandem source")
  else:
    return os.path.join(os.environ.get("PANDEM_HOME"), *path)

def settings():
  config = pandem_path("settings.yml")
  with open(config, "r") as f:
    settings = yaml.safe_load(f)
  if os.environ.get("PANDEM_NLP") is not None:
     settings["pandem"]["source"]["nlp"]["models_path"] = os.environ.get("PANDEM_NLP")
  return settings

def absolute_to_relative(path, inner_path):
    rel = os.environ.get("PANDEM_HOME")
    if path.startswith(rel):
      path = path[len(rel):len(path)]
    if path.startswith("/") or path.startswith("\\"):
      path = path[1:len(path)]
    rel = inner_path
    if path.startswith(rel):
      path = path[len(rel):len(path)]
    return path

def pretty(o):
  return json.dumps(o,cls=JsonEncoder, indent = 4)

def get_or_set_secret(name):
  secret_dir = pandem_path("secrets")
  if not os.path.exists(secret_dir):
    os.makedirs(name = secret_dir, mode = 0o700)

  secret_path = os.path.join(secret_dir, name)
  if not os.path.exists(secret_path):
    p = getpass.getpass(prompt = f"Password {name} not found, please type it or put a file with its content (UTF-8) in {secret_path}")
    with open(secret_path, 'w', encoding='utf-8') as f:
      f.write(p)
    os.chmod(secret_path, 0o400)
  with io.open(secret_path, mode = "r", encoding = "utf-8") as f:
    return f.read()

def get_custom(path, function):
  if type(path) == str:
    path = [path]
  path = [*( p.replace('-', '_') for p in path )]
  script_path = pandem_path(*(["files", "scripts", "py"] + path))
  if not os.path.exists(script_path+".py"):
    return None
  else :
    exec(f"import {'.'.join(path)}")
    if not eval(f"hasattr({'.'.join(path)}, '{function}')"):
      return None
    else:
      return eval(f"{'.'.join(path)}.{function}")


class JsonEncoder(json.JSONEncoder):
  def default(self, z):
    if isinstance(z, datetime.datetime) or isinstance(z, datetime.date):
      return (str(z))
    elif  isinstance(z, numpy.int64):
      return int(z)
    else:
      return super().default(z)




import os
import requests
import hashlib
from . import acquisition
from urllib.parse import urlparse
from urllib.parse import parse_qsl

class AcquisitionURL(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "url")
        
    def new_files(self, dls, last_hash):
        urls = dls['acquisition']['channel']['url']
        if type(urls) == str:
          urls = [urls]
        hashes = last_hash.split("|")
        pairs = [(urls[i], (hashes[i] if i<len(hashes) else '')) for i in range(0, len(urls))]
        source_dir = self.source_path(dls)
        current_hash = []
        files_to_pipeline = []
        for url, last_hash in pairs:
            # getting the file name
            parts = urlparse(url)
            if 'file_name' not in dls['acquisition']['channel']:
              dest_name = parts.path.replace("/", "_")
            elif type(dls['acquisition']['channel']['file_name'])==str :
              dest_name = dls['acquisition']['channel']['file_name']
            else:
              dest_name = ''
              for p in dls['acquisition']['channel']['file_name']:
                if p.startswith("query."):
                  for par, val in parse_qsl(parts.query):
                    if p == f"query.{par}":
                      dest_name = f"{dest_name}_{par}-{val}"
                elif p == "netloc":
                  dest_name = f"{dest_name}_{parts.netloc.replace('/', '_')}"
                elif p == "path":
                  dest_name = f"{dest_name}_{parts.path.replace('/', '_')}"
              if dest_name == '':
                dest_name = parts.path.replace("/", "_")

            file_path = self.source_path(dls, dest_name)
            r = requests.get(url)
            current_hash.append(r.headers.get('ETag') if r.headers.get('Etag') is not None else hashlib.sha1(r.content).hexdigest())
            if not os.path.exists(file_path) or last_hash == "":
                with open (file_path,'wb') as cont:
                    cont.write(r.content)
                files_to_pipeline.extend([file_path])
            # the file already exists and we know the last etag 
            elif current_hash[-1] != last_hash:
                with open (file_path,'wb') as cont:
                    cont.write(r.content)
                files_to_pipeline.extend([file_path])   
        return {"hash":"|".join(current_hash), "files":files_to_pipeline} 





import numpy as np
import pandas as pd

def df_transform(df):
  def ag(i):
    if pd.isna(df['age_min'][i]) and pd.isna(df['age_min'][i]):
      return pd.NA
    elif pd.isna(df['age_min'][i]):
      return "-"+str(int(df['age_max'][i]))
      pd.isna()+ "-" + df['age_max'].map(str) 
    elif pd.isna(df['age_max'][i]):
      return str(int(df['age_min'][i]))+"-"
    else:
      return str(int(df['age_min'][i])) + "-" + str(int(df['age_max'][i]))
  
  df['age_group'] = [*map(ag, df.index)]
  return df



from pandemsource import acquisition_zenodo
import pykka
from . import storage
from . import acquisition_url
from . import acquisition_git
from . import pipeline
from . import formatreader_xml
from . import formatreader_csv
from . import formatreader_xls
from . import formatreader_json
from . import unarchive
from . import acquisition_git_local
from . import acquisition_localFS
from . import acquisition_twitter
from . import acquisition_medisys
from . import script_executor
from . import nlp_annotator
from . import dfreader
from . import standardizer
from . import aggregator
from . import variables
from . import evaluator
from . import api
import datetime
import os
import pandas as pd
from itertools import chain

class Orchestration(pykka.ThreadingActor):

    def __init__(self, settings, start_acquisition = True, retry_failed = False, restart_job = 0, retry_active = True, force_acquire = True, no_nlp = False):
        super(Orchestration, self).__init__()
        self.settings = settings
        self.current_actors = dict()
        self.start_acquisition = start_acquisition
        self.retry_failed = retry_failed
        self.retry_active = retry_active
        self.restart_job = restart_job
        self.force_acquire = force_acquire
        self.no_nlp = no_nlp

    def on_start(self):
        # Launching the storage actor which can be used by any actor
        storage_ref = storage.Storage.start('storage', self.actor_ref, self.settings)
        storage_proxy = storage_ref.proxy()
        self.current_actors['storage'] = {'ref': storage_ref}

        # Launching passive actors
        # launch dataframe reader actor
        dfreader_ref = dfreader.DataframeReader.start('dfreader', self.actor_ref, storage_ref, self.settings)
        self.current_actors['dfreader'] = {'ref': dfreader_ref}
        
        #launch xml format reader actor
        xmlreader_ref = formatreader_xml.FormatReaderXML.start('ftreader_xml', self.actor_ref, self.settings)
        self.current_actors['ftreader_xml'] = {'ref': xmlreader_ref}

        #launch csv format reader actor
        csvreader_ref = formatreader_csv.FormatReaderCSV.start('ftreader_csv', self.actor_ref, self.settings)
        self.current_actors['ftreader_csv'] = {'ref': csvreader_ref}

        #launch xls format reader actor
        xlsreader_ref = formatreader_xls.FormatReaderXLS.start('ftreader_xls', self.actor_ref, self.settings)
        self.current_actors['ftreader_xls'] = {'ref': xlsreader_ref}

        #launch json format reader actor
        jsonreader_ref = formatreader_json.FormatReaderJSON.start('ftreader_json', self.actor_ref, self.settings)
        self.current_actors['ftreader_json'] = {'ref': jsonreader_ref}
        
        #launch unarchiver actor
        unarchive_ref = unarchive.Unarchive.start('unarchiver', self.actor_ref, self.settings)
        self.current_actors['unarchiver'] = {'ref': unarchive_ref}

        # launch pipeline actor
        pipeline_ref = pipeline.Pipeline.start('pipeline', self.actor_ref, self.settings, self.retry_failed, self.restart_job, self.retry_active)
        self.current_actors['pipeline'] = {'ref': pipeline_ref}

        # launch script executor reader
        script_executor_ref = script_executor.ScriptExecutor.start('script_executor', self.actor_ref, self.settings)
        self.current_actors['script_executor'] = {'ref': script_executor_ref}

        # launch nlp annotator actor
        run_nlp = self.settings["pandem"]["source"]["nlp"]["active"] and not self.no_nlp
        nlp_annotator_ref = nlp_annotator.NLPAnnotator.start('nlp_annotator', self.actor_ref, self.settings, run_nlp)
        self.current_actors['nlp_annotator'] = {'ref': nlp_annotator_ref}
        
        # launch standardizer actor
        standardizer_ref = standardizer.Standardizer.start('standardizer', self.actor_ref, self.settings)
        self.current_actors['standardizer'] = {'ref': standardizer_ref}

        # launch aggregator actor
        aggregator_ref = aggregator.Aggregator.start('aggregator', self.actor_ref, self.settings)
        self.current_actors['aggregator'] = {'ref': aggregator_ref}
        
        # launch variables actor
        variables_ref = variables.Variables.start('variables', self.actor_ref, self.settings)
        self.current_actors['variables'] = {'ref': variables_ref}

         # launch evaluator actor
        evaluator_ref = evaluator.Evaluator.start('evaluator', self.actor_ref, self.settings)
        self.current_actors['evaluator'] = {'ref': evaluator_ref}

         # launch api actor
        api_ref = api.apiREST.start('api', self.actor_ref, self.settings)
        self.current_actors['api'] = {'ref': api_ref}

        # Launching acquisition actors (active actors)
        # List source definition files within 'source-definitions' through storage actor to get 
        source_files = storage_proxy.list_files('source-definitions').get()
        # read json dls files into dicts
        dls_dicts = [storage_proxy.read_file(file_name['path']).get() for file_name in source_files if file_name['path'].endswith(".json")]
        sources_labels = set([dls['acquisition']['channel']['name'] for dls in dls_dicts 
          if run_nlp or len([c for c in dls['columns'] if 'variable' in c and c['variable'] == "article_text"]) == 0
        ])
        
        for label in sources_labels:
            if label == "url":
                acquisition_ref = acquisition_url.AcquisitionURL.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "git":
                acquisition_ref = acquisition_git.AcquisitionGIT.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "git-local":
                acquisition_ref = acquisition_git_local.AcquisitionGITLocal.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "input-local":
                acquisition_ref = acquisition_localFS.AcquisitionLocalFS.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "twitter":
                acquisition_ref = acquisition_twitter.AcquisitionTwitter.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "medisys":
                acquisition_ref = acquisition_medisys.AcquisitionMedisys.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            elif label == "zenodo":
                acquisition_ref = acquisition_zenodo.AcquisitionZenodo.start(name = 'acquisition_'+label, orchestrator_ref = self.actor_ref, settings = self.settings)
            else:
                raise NotImplementedError(f"The acquisition channel {label} has not been implemented")

            acquisition_proxy = acquisition_ref.proxy()
            dls_label = [dls for dls in dls_dicts if dls['acquisition']['channel']['name'] == label]
            for dls in dls_label:
                if self.start_acquisition:
                    acquisition_proxy.add_datasource(dls, self.force_acquire)
        
            self.current_actors['acquisition_'+label] = {'ref': acquisition_ref, 'sources': dls_label}


    def get_heartbeat(self, actor_name):
        now = datetime.datetime.now()
        self.current_actors[actor_name]['heartbeat'] = now
        #print(f'heartbeat from {actor_name} at: {now}')

    def get_actor(self, actor_name):
        return self.current_actors[actor_name]['ref']
        

    def actors_df(self):
      map = self.current_actors
      df = pd.DataFrame(data = {
        "actor": map.keys(),
        "last_seen_seconds": [((datetime.datetime.now() - it["heartbeat"]).total_seconds() if "heartbeat" in it else pd.NA)  for it in map.values()],
        "heartbeat": [(it["heartbeat"] if "heartbeat" in it else pd.NA)  for it in map.values()]
      })
      return df



import yaml
import os

class config:
  _settings = {}

  def __init__(self, location = None):
    # loading default config
    pkg_dir, this_filename = os.path.split(__file__)
    defaults = os.path.join(pkg_dir, "data/defaults.yml") 
    with open(defaults, "r") as f:
      conf = yaml.safe_load(f)   
    if location is not None and os.path.exists(location):
      with open(location, "r") as f:
        userconf = yaml.safe_load(f)
      merge_config(conf, userconf)
    if "pandem" in conf and "sources" in conf["pandem"]: 
      self._settings = conf["pandem"]["sources"]
    else: 
      raise KeyError("Cannot find attribute pandem.sources on configuration")

  def get(self, path):
    setting = self._settings
    for p in path.split("."):
      if type(setting) is dict and p in setting:
        setting = setting[p]
      else:
        raise KeyError(f"Cannot find property '{p}' of '{path}' on settings")
    return setting

  def find(self, paths = None):
    if paths == None:
      return flatten_map(self._settings)
    else:
      if type(paths) is str:
        paths = [paths]
      ret = {}
      for path in paths:
        settings = self._settings
        base_path = ""
        for p in path.split("."):
          base_path = f"{base_path}.{p}" if len(base_path)>0 else p 
          if type(settings) is dict and p in settings:
            settings = settings[p]
          else:
            raise KeyError(f"Cannot find property '{p}' of '{path}' on settings")
        if type(settings) is dict:
          flatten_map(settings, path = base_path, curr = ret)
        else:
          ret[base_path] = setting
      return ret

def merge_maps(a, b, path = []):
  "merges map b into a"
  if path is None: path = []
  for key in b:
    if key in a:
      if isinstance(a[key], dict) and isinstance(b[key], dict):
        merge_maps(a[key], b[key], path + [str(key)])
      elif a[key] == b[key]:
        pass # same leaf value
      else:
        raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
    else:
      a[key] = b[key]
  return a

def flatten_map(m, path = None, curr = {}):
  assert type(m) is dict, f"To flatten map {type(m)} was expected to be a dictionary"
  for key, value in m.items():
    next_path = f"{path}.{key}" if path is not None else key
    if type(value) is dict:
      flatten_map(value, path = next_path, curr = curr)
    else: 
      curr[next_path] = value
  return curr




from . import formatreader
import pandas as pd

class FormatReaderXLS(formatreader.FormatReader):

    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
    
    def read_df(self, file_path, dls):
        file_bytes = self._storage_proxy.read_file(file_path).get()
        if file_bytes != '':
            sheet = dls['acquisition']['channel']['sheet']
            col_names = set([col['name'] for col in dls['columns']])
            df = pd.read_excel(file_bytes, sheet, header=None)
            df = df.dropna(how='all').dropna(how='all', axis=1)
            cols_row = 0
            best_row = 0
            missing = col_names
            while cols_row < len(df) and len(missing) > 0: 
                row_missing = col_names.difference(df.iloc[cols_row])
                if len(row_missing) < len(missing):
                    best_row = cols_row
                    missing = row_missing
                cols_row += 1
            if len(missing) > 0:
              raise ValueError(f"Cannot find columns {missing} on sheet {sheet}")
            cols = df.iloc[best_row]
            df  = pd.DataFrame(df.values[best_row+1:], columns=cols)
        else:
            raise ValueError('can not read from empty bytes')
        return df




from . import formatreader
import pandas as pd
import re
from lxml import etree

class FormatReaderXML(formatreader.FormatReader):

    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)

    def read_df(self, file_path, dls):
        tree = etree.parse(file_path)
        root = tree.getroot()
        nsmap = root.nsmap
        cols = [col['name'] for col in dls['columns'] ]
        rows = []
        for row in root.xpath(dls['acquisition']['format']['row'], namespaces=nsmap):
            dict_col = dict()
            for col in dls['columns']:
              if 'xpath' in col: 
                ltext = row.xpath(col['xpath'], namespaces=nsmap)
                if ltext: 
                    if 'find' in col:
                        subtext = [re.findall( col['find'], text)[0] for text in ltext]
                        if len(subtext)==1:
                            dict_col[col['name']] = subtext[0]
                        else:
                            dict_col[col['name']] = subtext
                    else:
                        dict_col[col['name']] = ltext[0]
                else:
                    dict_col[col['name']] = None
            rows.append(dict_col)
        df = pd.DataFrame(rows, columns = cols)
        return df


          



import pandas as pd

def df_transform(df):
  filtered_df = df.groupby(['day', 'destination'])['callsign'].count()
  filtered_df = filtered_df.reset_index()
  filtered_df['line_number'] = range(1, len(filtered_df)+1) 
  return filtered_df.rename(columns={"callsign": "flights"})


import os
import re
import threading
from . import acquisition
import subprocess
import re

class AcquisitionGIT(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "git")

    def new_files(self, dls, last_hash):
        #print(f'Getting new files wih hash {last_hash}')
        last_hash = last_hash.strip() 
        if("url" in dls['acquisition']['channel']):
          repo_name = dls['acquisition']['channel']['url'].split('/')[-1].split(".")[0]
          repo_dir = self.source_path(dls, repo_name)
        else :
          repo_name = ""
          repo_dir = self.source_path(dls)

        source_dir = self.source_path(dls)
        if "branch" in dls['acquisition']['channel']:
            dist_branch = dls['acquisition']['channel']['branch']
        else :
            dist_branch = "master"

        # If the repository does not exists or if no commit is provided all files will be sent to the pipeline
        files_to_pipeline = []
        if not os.path.exists(repo_dir) or last_hash == "":
            # clonning repo is a repo remote url has been provided
            if("url" in dls['acquisition']['channel']):
              subprocess.run(['git', 'clone',  dls['acquisition']['channel']['url']], cwd=source_dir) 
            
            #send all repo files in target subdirectories to the pipeline actor
            if "paths" in dls['acquisition']['channel']:
              subdirs = dls['acquisition']['channel']['paths']
            else :
              subdirs = [""]

            for subdir in subdirs:
              files_paths = self._storage_proxy.list_files(self.source_path(dls, repo_name, subdir)).get()
              if "match" in dls['acquisition']['channel']:
                files_paths = [f for f in files_paths if re.match(".*"+dls['acquisition']['channel']['match']+".*", f['name']) is not None]
              files_to_pipeline.extend([file_path['path'] for file_path in files_paths ])

        # the repo already exists and we know the last commit 
        else :             
            if("url" in dls['acquisition']['channel']):
                subprocess.run(['git', 'pull',  'origin', dist_branch], cwd = repo_dir)
            
            #get updated files
            if "paths" in dls['acquisition']['channel']:
              subdirs = dls['acquisition']['channel']['paths']
            else:
              subdirs = ['']
            
            for subdir in subdirs:
                cmd = ['git', 'diff', '--name-only', last_hash, 'HEAD']
                if subdir != '':
                  cmd.extend(subdir)
                new_files_subdir = subprocess.run(cmd, 
                                            capture_output=True,
                                            text=True,
                                            cwd=repo_dir
                )
                if new_files_subdir.stdout != '':
                    new_files = new_files_subdir.stdout.rstrip().split('\n')
                    files_paths =  [self.source_path(dls, repo_name, new_file) for new_file in new_files]
                    if "match" in dls['acquisition']['channel']:
                       files_paths = [f for f in files_paths if re.match(".*"+dls['acquisition']['channel']['match']+".*", f) is not None]
                    files_to_pipeline.extend(files_paths)
            
        new_commit = subprocess.run(['git', 'rev-parse', dist_branch], 
                        capture_output=True,
                        text=True,
                        cwd=repo_dir 
        ).stdout.strip()   
        print(f"{len(files_to_pipeline)} new files found for {source_dir}")
        ret = {"hash":new_commit, "files":files_to_pipeline}
        return ret



import subprocess
import os
import time
import threading
from . import worker
from . import util
from datetime import datetime, timedelta
import copy
import logging as l


class Standardizer(worker.Worker):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)

    def on_start(self):
        super().on_start()
        self._storage_proxy=self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._variables_proxy=self._orchestrator_proxy.get_actor('variables').get().proxy()
        self._pipeline_proxy=self._orchestrator_proxy.get_actor('pipeline').get().proxy()

    #def standardize(self, tuples_to_validate, job):
    def standardize(self, tuples, path, job, dls):  
        """
            IN:         tuples_to_validate and object job
            ACTIONS:    check the code and updates with a code if the values are in a ref
            OUT:        if issues not null list of issue
                        else list of tuples standardize
        """
        # getting tuples from cache
        tuples = tuples.value()

        std_tuples={'scope':{}, 'tuples':[]}
        std_var={}
        variables=self._variables_proxy.get_variables().get()
        list_issues=[]
        list_ref=[]
        global_tuple={}
        update_tuple={}
        refs_alias={}
        refs_values={}
        ignore_check = set(c["variable"] for c in dls["columns"] if "action" in c and c["action"] == "insert")
        type_translate=['referential_alias', 'referential_label']
        type_validate=['referential', 'geo_referential', 'referential_alias', 'referential_label']
        ref_matched = {}
        ref_failed = {}
        for i in range(-2, len(tuples['tuples'])):
            std_var = {}
            #retrieves the globals variable
            if i == -2: 
                std_var['attrs']=dict([(x['variable'],x['value']) for x in tuples['scope']['globals'] if 'value' in x])
            #retrieves the update variable
            elif i == -1: 
                std_var['attrs']=dict([(x['variable'],x['value']) for x in tuples['scope']['update_scope']])
            #retrieves the tupple
            else:
                std_var=copy.deepcopy(tuples['tuples'][i])
            validation_failed = False
            for var_name in std_var['attrs'].copy().keys():
                if variables[var_name]['type'] in type_translate:
                  code=variables[var_name]['linked_attributes'][0]
                else:
                  code = None
                #retrieves the referentiel
                if var_name not in refs_values and var_name in variables and variables[var_name]['type'] in type_validate:
                    referential=self._variables_proxy.get_referential(var_name).get()
                    if referential is not None:
                        refs_values[var_name]=set([x['attr'][var_name] for x in referential])
                        ref_matched[var_name] = False
                        ref_failed[var_name] = False
                    elif var_name not in ignore_check : 
                        self._pipeline_proxy.standardize_end(tuples = None, n_tuples = 0, issues = [self.nothing_found_issue(tuples['scope']['file_name'], job, var_name)], path = path, job = job)
                        return
                if var_name not in refs_alias and var_name in variables and variables[var_name]['type'] in type_translate: 
                    alias=self._variables_proxy.get_referential(var_name).get() 
                    if alias is not None:
                        refs_alias[var_name] = dict((x['attr'][var_name],x['attrs'][code]) for x in alias if code in x['attrs'])
                        ref_matched[var_name] = False
                        ref_failed[var_name] = False
                    elif var_name not in ignore_check : 
                        self._pipeline_proxy.standardize_end(tuples = None, n_tuples = 0, issues = [self.nothing_found_issue(tuples['scope']['file_name'], job, var_name)], path = path, job = job)
                        return

                var_type = variables[var_name]['type']
                #variable type is referentiel_translate
                if type(std_var['attrs'][var_name]) == list:
                  values = std_var['attrs'][var_name]
                  islist = True
                else: 
                  values = [std_var['attrs'][var_name]] 
                  islist = False
                if var_name in ignore_check or var_name not in variables:
                  pass
                #elif variables[var_name]['type'] in (type_validate + type_translate) and refs_values[var_name] is None:
                elif var_type in type_translate + type_validate :
                  new_values = []
                  var_ref = refs_values[var_name]
                  if var_type in type_translate:
                    std_var['attrs'].pop(var_name)
                  for var_value in values:
                    if var_value is not None and var_value in var_ref:
                      ref_matched[var_name] = True
                      if var_type in type_translate:
                        new_values.append(refs_alias[var_name][var_value])
                      else: 
                        new_values.append(var_value)
                    elif var_value is not None:
                      #Create a issue since validation failed 
                      ref_failed[var_name] = True
                      file_name = tuples['scope']['file_name']
                      if i >= 0:
                        line_number = tuples['tuples'][i]['attrs']['line_number']
                      else:
                        line_number = -1
                      message=(f"Code {var_value} does not exist in referential '{var_name}'. Line {line_number} in file {file_name} (source: '{tuples['scope']['source']}').")
                      issue={ "step":job['step'], 
                              "line":line_number, 
                              "source":tuples['scope']['source'], 
                              "file":file_name, 
                              "message":message, 
                              "raised_on":datetime.now(), 
                              "job_id":job['id'], 
                              "issue_type":"ref-not-found",
                              'issue_severity':"warning"
                      }
                      validation_failed = True
                      list_issues.append(issue)
                  if len(new_values)>0:
                    if var_type in type_translate:
                      dest_var = code
                    else:
                      dest_var = var_name
                    if islist:
                      std_var['attrs'][dest_var]=new_values
                    else:
                      std_var['attrs'][dest_var]=new_values[0]
            if i == -2: 
                global_tuple=std_var
            elif i == -1: 
                update_tuple=std_var
            else:
                for cle, value in global_tuple['attrs'].items():
                    if cle not in std_var['attrs']:
                       std_var['attrs'][cle]=value
                if not validation_failed:
                  std_tuples['tuples'].append(std_var)
        # the source will be delayed if there referentials with some failures and no success
        for ref, failed in ref_failed.items():
          if failed and not ref_matched[ref]:
            self._pipeline_proxy.standardize_end(tuples = None, n_tuples = 0, issues = list_issues, path = path, job = job)
            return
        std_tuples['scope']['update_scope']= [*({'variable':k, 'value':v} for k,v in update_tuple['attrs'].items())]
        #print("\n".join(util.pretty(std_tuples).split("\n")[0:100]))
        
        ret = self._storage_proxy.to_job_cache(job["id"], f"std_{path}", std_tuples).get()
        self._pipeline_proxy.standardize_end(tuples = ret, n_tuples = len(std_tuples['tuples']), issues = list_issues, path = path, job = job)

    def nothing_found_issue(self, file_name, job, var_name):
      return { "step":job['step'], 
        "line":0, 
        "source":job["source"], 
        "file":file_name, 
        "message":f"Cannot find any match for variable {var_name} on {file_name} this standardize will be retried if all files on source have the same issue", 
        "raised_on":datetime.now(), 
        "job_id":job['id'], 
        "issue_type":"ref-not-found",
        'issue_severity':"warning"
      }
